# Copyright (c) 2013 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys

# NOTE(arosen): done to avoid the fact that cloudfoundryclient
# isn't in the openstack global reqirements.
import mock
sys.modules['cloudfoundryclient.v2.client'] = mock.Mock()
sys.modules['cloudfoundryclient.v2'] = mock.Mock()
sys.modules['cloudfoundryclient'] = mock.Mock()

from congress.datasources import cloudfoundryv2_driver
from congress.tests import base
from congress.tests import helper


ORG1_GUID = '5187136c-ef7d-47e6-9e6b-ac7780bab3db'
ORG_DATA = (
    {"total_results": 1,
     "next_url": 'null',
     "total_pages": 1,
     "prev_url": 'null',
     "resources": [{
         "entity":
             {"status": "active",
              "spaces_url": "/v2/organizations/" + ORG1_GUID + "/spaces",
              "private_domains_url":
                 "/v2/organizations/" + ORG1_GUID + "/private_domains",
              "name": "foo.com",
              "domains_url":
                 "/v2/organizations/" + ORG1_GUID + "/domains",
              "billing_enabled": 'true',
              "quota_definition_guid":
                 "b72b1acb-ff4f-468d-99c0-05cd91012b62",
              "app_events_url":
                 "/v2/organizations/" + ORG1_GUID + "/app_events",
              "space_quota_definitions_url":
                 "/v2/organizations/" + ORG1_GUID + "/space_quota_definitions",
              "quota_definition_url":
                 "/v2/quota_definitions/b72b1acb-ff4f-468d-99c0-05cd91012b62",
              "auditors_url":
                 "/v2/organizations/" + ORG1_GUID + "/auditors",
              "managers_url":
                 "/v2/organizations/" + ORG1_GUID + "/managers",
              "users_url":
                 "/v2/organizations/" + ORG1_GUID + "/users",
              "billing_managers_url":
                 "/v2/organizations/" + ORG1_GUID + "/billing_managers"
              },
         "metadata":
             {"url":
                 "/v2/organizations/5187136c-ef7d-47e6-9e6b-ac7780bab3db",
              "created_at": "2015-01-21T02:17:28+00:00",
              "guid": "5187136c-ef7d-47e6-9e6b-ac7780bab3db",
              "updated_at": "2015-01-21T02:17:28+00:00"
              }
     }
     ]
     }
)


SPACE1_GUID = "8da5477d-340e-4bb4-808a-54d9f72017d1"
SPACE2_GUID = "79479021-1e77-473a-8c63-28de9d2ca697"
ORG1_SPACES_DATA = (
    {"total_results": 2,
     "next_url": "null",
     "total_pages": 1,
     "prev_url": "null",
     "resources": [{
         "entity":
             {"developers_url": "/v2/spaces/" + SPACE1_GUID + "/developers",
              "service_instances_url":
                 "/v2/spaces/" + SPACE1_GUID + "/service_instances",
              "events_url": "/v2/spaces/" + SPACE1_GUID + "/events",
              "name": "development",
              "domains_url": "/v2/spaces/" + SPACE1_GUID + "/domains",
              "app_events_url": "/v2/spaces/" + SPACE1_GUID + "/app_events",
              "routes_url": "/v2/spaces/" + SPACE1_GUID + "/routes",
              "organization_guid": "5187136c-ef7d-47e6-9e6b-ac7780bab3db",
              "space_quota_definition_guid": "null",
              "apps_url": "/v2/spaces/" + SPACE1_GUID + "/apps",
              "auditors_url": "/v2/spaces/" + SPACE1_GUID + "/auditors",
              "managers_url": "/v2/spaces/" + SPACE1_GUID + "/managers",
              "organization_url":
                 "/v2/organizations/5187136c-ef7d-47e6-9e6b-ac7780bab3db",
              "security_groups_url":
                 "/v2/spaces/" + SPACE1_GUID + "/security_groups"
              },
         "metadata":
             {"url": "/v2/spaces/" + SPACE1_GUID,
              "created_at": "2015-01-21T02:17:28+00:00",
              "guid": SPACE1_GUID,
              "updated_at": "null"
              }
         },
         {"entity":
             {"developers_url": "/v2/spaces/" + SPACE2_GUID + "/developers",
              "service_instances_url":
                 "/v2/spaces/" + SPACE2_GUID + "/service_instances",
              "events_url": "/v2/spaces/" + SPACE2_GUID + "/events",
              "name": "test2",
              "domains_url": "/v2/spaces/" + SPACE2_GUID + "/domains",
              "app_events_url": "/v2/spaces/" + SPACE2_GUID + "/app_events",
              "routes_url": "/v2/spaces/" + SPACE2_GUID + "/routes",
              "organization_guid": "5187136c-ef7d-47e6-9e6b-ac7780bab3db",
              "space_quota_definition_guid": "null",
              "apps_url": "/v2/spaces/" + SPACE2_GUID + "/apps",
              "auditors_url": "/v2/spaces/" + SPACE2_GUID + "/auditors",
              "managers_url": "/v2/spaces/" + SPACE2_GUID + "/managers",
              "organization_url":
                 "/v2/organizations/5187136c-ef7d-47e6-9e6b-ac7780bab3db",
              "security_groups_url":
                 "/v2/spaces/" + SPACE2_GUID + "/security_groups"
              },
          "metadata":
              {"url": "/v2/spaces/" + SPACE2_GUID,
               "created_at": "2015-01-22T19:02:32+00:00",
               "guid": SPACE2_GUID,
               "updated_at": "null"
               }
          }
         ]
     }
)

APP1_GUID = "c3bd7fc1-73b4-4cc7-a6c8-9976c30edad5"
APP2_GUID = "f7039cca-95ac-49a6-b116-e32a53ddda69"
APPS_IN_SPACE1 = (
    {"total_results": 2,
     "next_url": "null",
     "total_pages": 1,
     "prev_url": "null",
     "resources": [{
         "entity":
             {"version": "fec00ce7-a980-49e1-abec-beed5516618f",
              "staging_failed_reason": "null",
              "instances": 1,
              "routes_url": "/v2/apps" + APP1_GUID + "routes",
              "space_url": "/v2/spaces/8da5477d-340e-4bb4-808a-54d9f72017d1",
              "docker_image": "null",
              "console": "false",
              "package_state": "STAGED",
              "state": "STARTED",
              "production": "false",
              "detected_buildpack": "Ruby",
              "memory": 256,
              "package_updated_at": "2015-01-21T21:00:40+00:00",
              "staging_task_id": "71f75ad3cad64884a92c4e7738eaae16",
              "buildpack": "null",
              "stack_url": "/v2/stacks/50688ae5-9bfc-4bf6-a4bf-caadb21a32c6",
              "events_url": "/v2/apps" + APP1_GUID + "events",
              "service_bindings_url":
                 "/v2/apps" + APP1_GUID + "service_bindings",
              "detected_start_command":
                 "bundle exec rake db:migrate && bundle exec rails s -p $PORT",
              "disk_quota": 1024,
              "stack_guid": "50688ae5-9bfc-4bf6-a4bf-caadb21a32c6",
              "space_guid": "8da5477d-340e-4bb4-808a-54d9f72017d1",
              "name": "rails_sample_app",
              "health_check_type": "port",
              "command":
                 "bundle exec rake db:migrate && bundle exec rails s -p $PORT",
              "debug": "null",
              "environment_json": "null",
              "health_check_timeout": "null"
              },
         "metadata":
             {"url": "/v2/apps/c3bd7fc1-73b4-4cc7-a6c8-9976c30edad5",
              "created_at": "2015-01-21T21:01:19+00:00",
              "guid": "c3bd7fc1-73b4-4cc7-a6c8-9976c30edad5",
              "updated_at": "2015-01-21T21:01:19+00:00"
              }
         },
         {"entity":
             {"version": "a1b52559-32f3-4765-9fd3-6e35293fb6d0",
              "staging_failed_reason": "null",
              "instances": 1,
              "routes_url": "/v2/apps" + APP2_GUID + "routes",
              "space_url": "/v2/spaces/8da5477d-340e-4bb4-808a-54d9f72017d1",
              "docker_image": "null",
              "console": "false",
              "package_state": "PENDING",
              "state": "STOPPED",
              "production": "false",
              "detected_buildpack": "null",
              "memory": 1024,
              "package_updated_at": "null",
              "staging_task_id": "null",
              "buildpack": "null",
              "stack_url": "/v2/stacks/50688ae5-9bfc-4bf6-a4bf-caadb21a32c6",
              "events_url": "/v2/apps" + APP2_GUID + "events",
              "service_bindings_url":
                 "/v2/apps" + APP2_GUID + "service_bindings",
              "detected_start_command": "",
              "disk_quota": 1024,
              "stack_guid": "50688ae5-9bfc-4bf6-a4bf-caadb21a32c6",
              "space_guid": "8da5477d-340e-4bb4-808a-54d9f72017d1",
              "name": "help",
              "health_check_type": "port",
              "command": "null",
              "debug": "null",
              "environment_json": "null",
              "health_check_timeout": "null"
              },
          "metadata":
             {"url": "/v2/apps/f7039cca-95ac-49a6-b116-e32a53ddda69",
              "created_at": "2015-01-21T18:48:34+00:00",
              "guid": "f7039cca-95ac-49a6-b116-e32a53ddda69",
              "updated_at": "null"
              }
          }
         ]
     }
)


APPS_IN_SPACE2 = {"total_results": 0,
                  "next_url": "null",
                  "total_pages": 1,
                  "prev_url": "null",
                  "resources": []}

SERVICES_IN_SPACE1 = {
    "guid": "8da5477d-340e-4bb4-808a-54d9f72017d1",
    "name": "development",
    "services": [{
        "bound_app_count": 0,
        "guid": "88f61682-d78e-410f-88ee-1e0eabbbc7da",
        "last_operation": None,
        "name": "rails-postgres",
        "service_plan": {
            "guid": "fbcec3af-3e8d-4ee7-adfe-3f12a137ed66",
            "name": "turtle",
            "service": {
                "guid": "34dbc753-34ed-4cf1-9a87-a224dfca569b",
                "label": "elephantsql",
                "provider": None,
                "version": None
            }
        }
    }]
}

EXPECTED_STATE = {
    'organizations': set([
        ('5187136c-ef7d-47e6-9e6b-ac7780bab3db', 'foo.com',
         '2015-01-21T02:17:28+00:00', '2015-01-21T02:17:28+00:00')]),
    'spaces': set([
        ('8da5477d-340e-4bb4-808a-54d9f72017d1', 'development',
         '2015-01-21T02:17:28+00:00', 'null'),
        ('79479021-1e77-473a-8c63-28de9d2ca697', 'test2',
         '2015-01-22T19:02:32+00:00', 'null')]),
    'apps': set([
        ('8da5477d-340e-4bb4-808a-54d9f72017d1',
         'c3bd7fc1-73b4-4cc7-a6c8-9976c30edad5', 'null',
         'bundle exec rake db:migrate && bundle exec rails s -p $PORT',
         'false', 'null', 'Ruby',
         'bundle exec rake db:migrate && bundle exec rails s -p $PORT',
         1024, 'null', 'null', 'null', 1,
         256, 'rails_sample_app', 'STAGED', '2015-01-21T21:00:40+00:00',
         'false', 'null', '71f75ad3cad64884a92c4e7738eaae16', 'STARTED',
         'fec00ce7-a980-49e1-abec-beed5516618f', '2015-01-21T21:01:19+00:00',
         '2015-01-21T21:01:19+00:00'),
        ('8da5477d-340e-4bb4-808a-54d9f72017d1',
         'f7039cca-95ac-49a6-b116-e32a53ddda69', 'null', 'null', 'false',
         'null', 'null', '', 1024, 'null', 'null', 'null', 1, 1024,
         'help', 'PENDING', 'null', 'false', 'null', 'null', 'STOPPED',
         'a1b52559-32f3-4765-9fd3-6e35293fb6d0',
         '2015-01-21T18:48:34+00:00', 'null')]),
    'service_bindings': set([]),
    'services': set([
        ('88f61682-d78e-410f-88ee-1e0eabbbc7da',
         '8da5477d-340e-4bb4-808a-54d9f72017d1', 'rails-postgres',
         0, 'None', 'turtle')]),
}


class TestCloudFoundryV2Driver(base.TestCase):

    def setUp(self):
        super(TestCloudFoundryV2Driver, self).setUp()

        args = helper.datasource_openstack_args()
        args['poll_time'] = 0
        args['client'] = mock.MagicMock()
        self.driver = cloudfoundryv2_driver.CloudFoundryV2Driver(args=args)

    def test_update_from_datasource(self):
        def _side_effect_get_org_spaces(org):
            if org == ORG1_GUID:
                return ORG1_SPACES_DATA
            raise ValueError("This should occur...")

        def _side_effect_get_apps_in_space(space):
            if space == SPACE1_GUID:
                return APPS_IN_SPACE1
            elif space == SPACE2_GUID:
                return APPS_IN_SPACE2
            else:
                raise ValueError("This should not occur....")

        def _side_effect_get_spaces_summary(space):
            if space == SPACE1_GUID:
                return SERVICES_IN_SPACE1
            else:
                return {"guid": space,
                        "services": []}

        def _side_effect_get_app_services(space):
            return {'resources': []}

        with base.nested(
            mock.patch.object(self.driver.cloudfoundry,
                              "get_organizations",
                              return_value=ORG_DATA),
            mock.patch.object(self.driver.cloudfoundry,
                              "get_organization_spaces",
                              side_effect=_side_effect_get_org_spaces),
            mock.patch.object(self.driver.cloudfoundry,
                              "get_apps_in_space",
                              side_effect=_side_effect_get_apps_in_space),
            mock.patch.object(self.driver.cloudfoundry,
                              "get_spaces_summary",
                              side_effect=_side_effect_get_spaces_summary),
            mock.patch.object(self.driver.cloudfoundry,
                              "get_app_service_bindings",
                              side_effect=_side_effect_get_app_services),


            ) as (get_organizations, get_organization_spaces,
                  get_apps_in_space, get_spaces_summary,
                  get_app_services_guids):
            self.driver.update_from_datasource()
            self.assertEqual(EXPECTED_STATE, self.driver.state)

    def test_execute(self):
        class CloudfoundryClient(object):
            def __init__(self):
                self.testkey = None

            def setServices(self, arg1):
                self.testkey = 'arg1=%s' % arg1

        cloudfoundry_client = CloudfoundryClient()
        self.driver.cloudfoundry = cloudfoundry_client
        api_args = {
            'positional': ['1']
        }
        expected_ans = 'arg1=1'

        self.driver.execute('setServices', api_args)

        self.assertEqual(expected_ans, cloudfoundry_client.testkey)



# Copyright (c) 2013 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import json
import os
import time

from oslo_config import cfg
from oslo_log import log as logging
from oslo_messaging import conffixture
import retrying

from congress.datalog import compile
from congress.datalog import unify
from congress.policy_engines import agnostic

from congress.dse2.dse_node import DseNode


LOG = logging.getLogger(__name__)

ROOTDIR = os.path.dirname(__file__)
ETCDIR = os.path.join(ROOTDIR, 'etc')

# single, global variable used to ensure different tests from
#  different subclasses of TestCase all can get a unique ID
#  so that the tests do not interact on oslo-messaging
partition_counter = 0


def make_dsenode_new_partition(node_id,
                               messaging_config=None,
                               node_rpc_endpoints=None):
    """Get new DseNode in it's own new DSE partition."""
    messaging_config = messaging_config or generate_messaging_config()
    node_rpc_endpoints = node_rpc_endpoints or []
    return DseNode(messaging_config, node_id, node_rpc_endpoints,
                   partition_id=get_new_partition())


def make_dsenode_same_partition(existing,
                                node_id,
                                messaging_config=None,
                                node_rpc_endpoints=None):
    """Get new DseNode in the same DSE partition as existing (node or part)."""
    partition_id = (existing.partition_id if
                    isinstance(existing, DseNode) else existing)

    messaging_config = messaging_config or generate_messaging_config()
    node_rpc_endpoints = node_rpc_endpoints or []
    return DseNode(messaging_config, node_id, node_rpc_endpoints, partition_id)


def get_new_partition():
    """Create a new partition number, unique within each process."""
    global partition_counter
    old = partition_counter
    partition_counter += 1
    return old


def generate_messaging_config():
    mc_fixture = conffixture.ConfFixture(cfg.CONF)
    mc_fixture.conf.transport_url = 'kombu+memory://'
    messaging_config = mc_fixture.conf
    messaging_config.rpc_response_timeout = 1
    return messaging_config


def etcdir(*p):
    return os.path.join(ETCDIR, *p)


def root_path():
    """Return path to root of source code."""
    x = os.path.realpath(__file__)
    x, y = os.path.split(x)  # drop "helper.py"
    x, y = os.path.split(x)  # drop "tests"
    x, y = os.path.split(x)  # drop "congress"
    return x


def source_path():
    """Return path to root of source code."""
    x = os.path.realpath(__file__)
    x, y = os.path.split(x)  # drop "helper.py"
    x, y = os.path.split(x)  # drop "tests"
    return x


def data_module_path(file):
    """Return path to dataservice module with given FILEname."""
    path = source_path()
    path = os.path.join(path, "datasources")
    path = os.path.join(path, file)
    return path


def policy_module_path():
    """Return path to policy engine module."""
    path = source_path()
    path = os.path.join(path, "policy_engines")
    path = os.path.join(path, "agnostic.py")
    return path


def api_module_path():
    """Return path to api module."""
    path = source_path()
    path = os.path.join(path, "datasources")
    path = os.path.join(path, "test_driver.py")
    return path


def test_path():
    """Return path to root of top-level tests."""
    path = source_path()
    path = os.path.join(path, "tests")
    return path


def datasource_config_path():
    """Return path to configuration info for datasources."""
    path = test_path()
    path = os.path.join(path, "datasources.conf")
    return path


def datasource_openstack_args():
    """Return basic args for creating an openstack datasource."""
    return {'username': '',
            'password': '',
            'auth_url': '',
            'tenant_name': ''}


def pause(factor=1):
    """Timeout so other threads can run."""
    time.sleep(factor * 1)


def datalog_same(actual_code, correct_code, msg=None):
    return datalog_equal(
        actual_code, correct_code, msg=msg,
        equal=lambda x, y: unify.same(x, y) is not None)


def datalog_equal(actual_code, correct_code,
                  msg=None, equal=None, theories=None,
                  output_diff=True):
    """Check equality.

    Check if the strings given by actual_code
    and CORRECT_CODE represent the same datalog.
    """
    def minus(iter1, iter2, invert=False):
        extra = []
        for i1 in iter1:
            found = False
            for i2 in iter2:
                # for asymmetric equality checks
                if invert:
                    test_result = equal(i2, i1)
                else:
                    test_result = equal(i1, i2)
                if test_result:
                    found = True
                    break
            if not found:
                extra.append(i1)
        return extra
    if equal is None:
        equal = lambda x, y: x == y

    LOG.debug("** Checking equality: %s **", msg)
    actual = compile.parse(actual_code, theories=theories)
    correct = compile.parse(correct_code, theories=theories)
    extra = minus(actual, correct)
    # in case EQUAL is asymmetric, always supply actual as the first arg
    #   and set INVERT to true
    missing = minus(correct, actual, invert=True)
    if output_diff:
        output_diffs(extra, missing, msg)
    LOG.debug("** Finished equality: %s **", msg)
    is_equal = len(extra) == 0 and len(missing) == 0
    if not is_equal:
        LOG.debug('datalog_equal failed, extras: %s, missing: %s', extra,
                  missing)
    return is_equal


def db_equal(actual_string, correct_string, output_diff=True):
    """Check if two strings representing data theories are the same."""
    actual = agnostic.string_to_database(actual_string)
    correct = agnostic.string_to_database(correct_string)
    return check_db_diffs(actual, correct, output_diff=output_diff)


def check_db_diffs(actual, correct, output_diff=True):
    extra = actual - correct
    missing = correct - actual
    extra = [e for e in extra if not e[0].startswith("___")]
    missing = [m for m in missing if not m[0].startswith("___")]
    if output_diff:
        output_diffs(extra, missing, actual=actual)
    return len(extra) == 0 and len(missing) == 0


def output_diffs(extra, missing, actual=None):
    if len(extra) > 0:
        print("Extra tuples")
        print(", ".join([str(x) for x in extra]))
    if len(missing) > 0:
        print("Missing tuples")
        print(", ".join([str(x) for x in missing]))
    if len(extra) > 0 or len(missing) > 0:
        print("Resulting database: {}".format(str(actual)))


def str2form(formula_string, theories=None):
    return compile.parse1(formula_string, theories=theories)


def str2pol(policy_string, theories=None):
    return compile.parse(policy_string, theories=theories)


def pol2str(policy):
    return " ".join(str(x) for x in policy)


def form2str(formula):
    return str(formula)


@retrying.retry(stop_max_attempt_number=1000, wait_fixed=100)
def retry_check_for_last_message(obj):
    if not hasattr(obj, "last_msg"):
        raise AttributeError("Missing 'last_msg' attribute")


@retrying.retry(stop_max_attempt_number=1000, wait_fixed=100)
def retry_check_for_message_to_arrive(obj):
    if not hasattr(obj.msg, "body"):
        raise AttributeError("Missing 'body' attribute")


@retrying.retry(stop_max_attempt_number=1000, wait_fixed=100)
def retry_check_for_message_data(obj, data):
    if not hasattr(obj.msg, "body"):
        raise AttributeError("Missing 'body' attribute")
    if obj.get_msg_data() != data:
        raise Exception("Missing expected data in msg")


@retrying.retry(stop_max_attempt_number=1000, wait_fixed=100)
def retry_check_nonempty_last_policy_change(obj):
    if not hasattr(obj, "last_policy_change"):
        raise AttributeError("Missing 'last_policy_change' attribute")
    if obj.last_policy_change is None:
        raise Exception("last_policy_change == None")
    if len(obj.last_policy_change) == 0:
        raise Exception("last_policy_change == 0")


@retrying.retry(stop_max_attempt_number=1000, wait_fixed=100)
def retry_check_empty_last_policy_change(obj):
    if not hasattr(obj, "last_policy_change"):
        raise AttributeError("Missing 'last_policy_change' attribute")
    if len(obj.last_policy_change) != 0:
        raise Exception("last_policy_change != 0")


@retrying.retry(stop_max_attempt_number=1000, wait_fixed=100)
def retry_check_db_equal(policy, query, correct, target=None):
    if not hasattr(policy, "select"):
        raise AttributeError("Missing 'select' attribute")
    if target is None:
        actual = policy.select(query)
    else:
        actual = policy.select(query, target=target)
    if not db_equal(actual, correct, output_diff=False):
        raise Exception("Query {} produces {}, should produce {}".format(
            str(query), str(actual), str(correct)))


@retrying.retry(stop_max_attempt_number=1000, wait_fixed=100)
def retry_check_number_of_updates(deepsix, value):
    if not hasattr(deepsix, "number_of_updates"):
        raise AttributeError("Missing 'number_of_updates' attribute")
    if deepsix.number_of_updates != value:
        raise Exception("number_of_updates is {}, not {}".format(
            deepsix.number_of_updates, value))


@retrying.retry(stop_max_attempt_number=1000, wait_fixed=100)
def retry_check_subscriptions(deepsix, subscription_list):
    if not check_subscriptions(deepsix, subscription_list):
        raise Exception("{} does not have subscription list {}".format(
            deepsix.name, str(subscription_list)))


def check_subscriptions(deepsix, subscription_list):
    """Check subscriptions.

    Check that the instance DEEPSIX is subscribed to all of the
    (key, dataindex) pairs in KEY_DATAINDEX_LIST.  Return True if
    all subscriptions exists; otherwise returns False.
    """
    actual = set([(value.key, value.dataindex)
                  for value in deepsix.subdata.values()])
    correct = set(subscription_list)
    missing = correct - actual
    if missing:
        LOG.debug("Missing key/dataindex subscriptions: %s", missing)
    return not missing


@retrying.retry(stop_max_attempt_number=1000, wait_fixed=100)
def retry_check_subscribers(deepsix, subscriber_list):
    if not check_subscribers(deepsix, subscriber_list):
        raise Exception("{} does not have subscriber list {}".format(
            deepsix.name, str(subscriber_list)))


@retrying.retry(stop_max_attempt_number=1000, wait_fixed=100)
def retry_check_no_subscribers(deepsix, subscriber_list):
    if check_subscribers(deepsix, subscriber_list):
        raise Exception("{} still has subscriber list {}".format(
            deepsix.name, str(subscriber_list)))


def check_subscribers(deepsix, subscriber_list):
    """Check subscribers.

    Check that the instance DEEPSIX includes subscriptions for all of
    the (name, dataindex) pairs in SUBSCRIBER_LIST.  Return True if
    all subscribers exist; otherwise returns False.
    """
    actual = set([(name, pubdata.dataindex)
                  for pubdata in deepsix.pubdata.copy().values()
                  for name in pubdata.subscribers])
    correct = set(subscriber_list)
    missing = correct - actual
    if missing:
        LOG.debug("Missing name/dataindex subscribers: %s", missing)
    return not missing


@retrying.retry(stop_max_attempt_number=10, wait_fixed=500)
def retry_check_function_return_value(f, expected_value):
    """Check if function f returns expected key."""
    result = f()
    if result != expected_value:
        raise Exception("Expected value '%s' not received.  "
                        "Got %s instead." % (expected_value, result))


@retrying.retry(stop_max_attempt_number=10, wait_fixed=500)
def retry_check_function_return_value_not_eq(f, value):
    """Check if function f does not return expected value."""
    if f() == value:
        raise Exception("Actual value '%s' not different from '%s'" % value)


@retrying.retry(stop_max_attempt_number=10, wait_fixed=500)
def retry_til_exception(expected_exception, f):
    """Check if function f does not return expected value."""
    try:
        val = f()
        raise Exception("No exception thrown; received %s" % val)
    except expected_exception:
        return
    except Exception as e:
        raise Exception("Wrong exception thrown: %s" % e)


class FakeRequest(object):
    def __init__(self, body):
        self.body = json.dumps(body)


class FakeServiceObj(object):
    def __init__(self):
        self.state = {}



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import copy
import os
import os.path
import re
import sys

from oslo_config import cfg
from oslo_log import log as logging

from congress.datalog import base
from congress.dse import d6cage
from congress import exception
from congress.managers import datasource as datasource_manager


LOG = logging.getLogger(__name__)


def create(rootdir, config_override=None):
    """Get Congress up and running when src is installed in rootdir.

    i.e. ROOTDIR=/path/to/congress/congress.
    CONFIG_OVERRIDE is a dictionary of dictionaries with configuration
    values that overrides those provided in CONFIG_FILE.  The top-level
    dictionary has keys for the CONFIG_FILE sections, and the second-level
    dictionaries store values for that section.
    """
    LOG.debug("Starting Congress with rootdir=%s, config_override=%s",
              rootdir, config_override)

    # create message bus
    cage = d6cage.d6Cage()

    # read in datasource configurations

    cage.config = config_override or {}

    # path to congress source dir
    src_path = os.path.join(rootdir, "congress")

    datasource_mgr = datasource_manager.DataSourceManager()

    # add policy engine
    engine_path = os.path.join(src_path, "policy_engines/agnostic.py")
    LOG.info("main::start() engine_path: %s", engine_path)
    cage.loadModule("PolicyEngine", engine_path)
    cage.createservice(
        name="engine",
        moduleName="PolicyEngine",
        description="Policy Engine (DseRuntime instance)",
        args={'d6cage': cage, 'rootdir': src_path,
              'log_actions_only': cfg.CONF.enable_execute_action})
    engine = cage.service_object('engine')
    engine.initialize_table_subscriptions()
    engine.debug_mode()  # should take this out for production

    # add policy api
    api_path = os.path.join(src_path, "api/policy_model.py")
    LOG.info("main::start() api_path: %s", api_path)
    cage.loadModule("API-policy", api_path)
    cage.createservice(
        name="api-policy",
        moduleName="API-policy",
        description="API-policy DSE instance",
        args={'policy_engine': engine})

    # add rule api
    api_path = os.path.join(src_path, "api/rule_model.py")
    LOG.info("main::start() api_path: %s", api_path)
    cage.loadModule("API-rule", api_path)
    cage.createservice(
        name="api-rule",
        moduleName="API-rule",
        description="API-rule DSE instance",
        args={'policy_engine': engine})

    # add table api
    api_path = os.path.join(src_path, "api/table_model.py")
    LOG.info("main::start() api_path: %s", api_path)
    cage.loadModule("API-table", api_path)
    cage.createservice(
        name="api-table",
        moduleName="API-table",
        description="API-table DSE instance",
        args={'policy_engine': engine,
              'datasource_mgr': datasource_mgr})

    # add row api
    api_path = os.path.join(src_path, "api/row_model.py")
    LOG.info("main::start() api_path: %s", api_path)
    cage.loadModule("API-row", api_path)
    cage.createservice(
        name="api-row",
        moduleName="API-row",
        description="API-row DSE instance",
        args={'policy_engine': engine,
              'datasource_mgr': datasource_mgr})

    # add status api
    api_path = os.path.join(src_path, "api/status_model.py")
    LOG.info("main::start() api_path: %s", api_path)
    cage.loadModule("API-status", api_path)
    cage.createservice(
        name="api-status",
        moduleName="API-status",
        description="API-status DSE instance",
        args={'policy_engine': engine,
              'datasource_mgr': datasource_mgr})

    # add action api
    api_path = os.path.join(src_path, "api/action_model.py")
    LOG.info("main::start() api_path: %s", api_path)
    cage.loadModule("API-action", api_path)
    cage.createservice(
        name="api-action",
        moduleName="API-action",
        description="API-action DSE instance",
        args={'policy_engine': engine,
              'datasource_mgr': datasource_mgr})

    # add schema api
    api_path = os.path.join(src_path, "api/schema_model.py")
    LOG.info("main::start() api_path: %s", api_path)
    cage.loadModule("API-schema", api_path)
    cage.createservice(
        name="api-schema",
        moduleName="API-schema",
        description="API-schema DSE instance",
        args={'datasource_mgr': datasource_mgr})

    # add path for system/datasource-drivers
    api_path = os.path.join(src_path, "api/system/driver_model.py")
    LOG.info("main::start() api_path: %s", api_path)
    cage.loadModule("API-system", api_path)
    cage.createservice(
        name="api-system",
        moduleName="API-system",
        description="API-system DSE instance",
        args={'datasource_mgr': datasource_mgr})

    # Load policies from database
    engine.persistent_load_policies()

    # if this is the first time we are running Congress, need
    #   to create the default theories (which cannot be deleted)
    api_policy = cage.service_object('api-policy')

    engine.DEFAULT_THEORY = 'classification'
    engine.builtin_policy_names.add(engine.DEFAULT_THEORY)
    try:
        api_policy.add_item({'name': engine.DEFAULT_THEORY,
                             'description': 'default policy'}, {})
    except KeyError:
        pass

    engine.ACTION_THEORY = 'action'
    engine.builtin_policy_names.add(engine.ACTION_THEORY)
    try:
        api_policy.add_item({'kind': base.ACTION_POLICY_TYPE,
                             'name': engine.ACTION_THEORY,
                             'description': 'default action policy'},
                            {})
    except KeyError:
        pass

    # have policy-engine subscribe to api calls
    # TODO(thinrichs): either have API publish everything to DSE bus and
    #   have policy engine subscribe to all those messages
    #   OR have API interact with individual components directly
    #   and change all tests so that the policy engine does not need to be
    #   subscribed to 'policy-update'
    engine.subscribe('api-rule', 'policy-update',
                     callback=engine.receive_policy_update)

    # spin up all the configured services, if we have configured them

    drivers = datasource_mgr.get_datasources()
    # Setup cage.config as it previously done when it was loaded
    # from disk. FIXME(arosen) later!
    for driver in drivers:
        if not driver['enabled']:
            LOG.info("module %s not enabled, skip loading", driver['name'])
            continue
        driver_info = datasource_mgr.get_driver_info(driver['driver'])
        engine.create_policy(driver['name'], kind=base.DATASOURCE_POLICY_TYPE)
        try:
            cage.createservice(name=driver['name'],
                               moduleName=driver_info['module'],
                               args=driver['config'],
                               module_driver=True,
                               type_='datasource_driver',
                               id_=driver['id'])
        except d6cage.DataServiceError:
            # FIXME(arosen): If createservice raises congress-server
            # dies here. So we catch this exception so the server does
            # not die. We need to refactor the dse code so it just
            # keeps retrying the driver gracefully...
            continue
        service = cage.service_object(driver['name'])
        engine.set_schema(driver['name'], service.get_schema())

    # Insert rules.  Needs to be done after datasources are loaded
    #  so that we can compile away column references at read time.
    #  If datasources loaded after this, we don't have schemas.
    engine.persistent_load_rules()

    # Start datasource synchronizer after explicitly starting the
    # datasources, because the explicit call to create a datasource
    # will crash if the synchronizer creates the datasource first.
    synchronizer_path = os.path.join(src_path, "synchronizer.py")
    LOG.info("main::start() synchronizer: %s", synchronizer_path)
    cage.loadModule("Synchronizer", synchronizer_path)
    cage.createservice(
        name="synchronizer",
        moduleName="Synchronizer",
        description="DB synchronizer instance",
        args={'poll_time': cfg.CONF.datasource_sync_period})
    synchronizer = cage.service_object('synchronizer')
    engine.set_synchronizer(synchronizer)

    # add datasource api
    api_path = os.path.join(src_path, "api/datasource_model.py")
    LOG.info("main::start() api_path: %s", api_path)
    cage.loadModule("API-datasource", api_path)
    cage.createservice(
        name="api-datasource",
        moduleName="API-datasource",
        description="API-datasource DSE instance",
        args={'policy_engine': engine,
              'datasource_mgr': datasource_mgr,
              'synchronizer': synchronizer})

    return cage


def load_data_service(service_name, config, cage, rootdir, id_):
    """Load service.

    Load a service if not already loaded. Also loads its
    module if the module is not already loaded.  Returns None.
    SERVICE_NAME: name of service
    CONFIG: dictionary of configuration values
    CAGE: instance to load service into
    ROOTDIR: dir for start of module paths
    ID: UUID of the service.
    """
    config = copy.copy(config)
    if service_name in cage.services:
        return
    if 'module' not in config:
        raise exception.DataSourceConfigException(
            "Service %s config missing 'module' entry" % service_name)
    module_path = config['module']
    module_name = re.sub('[^a-zA-Z0-9_]', '_', module_path)
    if not os.path.isabs(module_path) and rootdir is not None:
        module_path = os.path.join(rootdir, module_path)
    if module_name not in sys.modules:
        LOG.info("Trying to create module %s from %s",
                 module_name, module_path)
        cage.loadModule(module_name, module_path)
    LOG.info("Trying to create service %s with module %s",
             service_name, module_name)
    cage.createservice(name=service_name, moduleName=module_name,
                       args=config, type_='datasource_driver', id_=id_)



# Copyright (c) 2013 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import novaclient.client
from oslo_log import log as logging
import six

from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils

LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    """This method is called by d6cage to create a dataservice instance."""
    return NovaDriver(name, keys, inbox, datapath, args)


class NovaDriver(datasource_driver.PollingDataSourceDriver,
                 datasource_driver.ExecutionDriver):
    SERVERS = "servers"
    FLAVORS = "flavors"
    HOSTS = "hosts"
    FLOATING_IPS = "floating_IPs"
    SERVICES = 'services'
    AVAILABILITY_ZONES = "availability_zones"

    # This is the most common per-value translator, so define it once here.
    value_trans = {'translation-type': 'VALUE'}

    def safe_id(x):
        if isinstance(x, six.string_types):
            return x
        try:
            return x['id']
        except Exception:
            return str(x)

    servers_translator = {
        'translation-type': 'HDICT',
        'table-name': SERVERS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'The UUID for the server',
              'translator': value_trans},
             {'fieldname': 'name', 'desc': 'Name of the server',
              'translator': value_trans},
             {'fieldname': 'hostId', 'col': 'host_id',
              'desc': 'The UUID for the host', 'translator': value_trans},
             {'fieldname': 'status', 'desc': 'The server status',
              'translator': value_trans},
             {'fieldname': 'tenant_id', 'desc': 'The tenant ID',
              'translator': value_trans},
             {'fieldname': 'user_id',
              'desc': 'The user ID of the user who owns the server',
              'translator': value_trans},
             {'fieldname': 'image', 'col': 'image_id',
              'desc': 'Name or ID of image',
              'translator': {'translation-type': 'VALUE',
                             'extract-fn': safe_id}},
             {'fieldname': 'flavor', 'col': 'flavor_id',
              'desc': 'Name of the flavor',
              'translator': {'translation-type': 'VALUE',
                             'extract-fn': safe_id}},
             {'fieldname': 'OS-EXT-AZ:availability_zone', 'col': 'zone',
              'desc': 'The availability zone of host',
              'translator': value_trans},
             {'fieldname': 'OS-EXT-SRV-ATTR:hypervisor_hostname',
              'desc': ('The hostname of hypervisor where the server is' +
                       'running'),
              'col': 'host_name', 'translator': value_trans})}

    flavors_translator = {
        'translation-type': 'HDICT',
        'table-name': FLAVORS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'desc': 'ID of the flavor',
              'translator': value_trans},
             {'fieldname': 'name', 'desc': 'Name of the flavor',
              'translator': value_trans},
             {'fieldname': 'vcpus', 'desc': 'Number of vcpus',
              'translator': value_trans},
             {'fieldname': 'ram', 'desc': 'Memory size in MB',
              'translator': value_trans},
             {'fieldname': 'disk', 'desc': 'Disk size in GB',
              'translator': value_trans},
             {'fieldname': 'ephemeral', 'desc': 'Ephemeral space size in GB',
              'translator': value_trans},
             {'fieldname': 'rxtx_factor', 'desc': 'RX/TX factor',
              'translator': value_trans})}

    hosts_translator = {
        'translation-type': 'HDICT',
        'table-name': HOSTS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'host_name', 'desc': 'Name of host',
              'translator': value_trans},
             {'fieldname': 'service', 'desc': 'Enabled service',
              'translator': value_trans},
             {'fieldname': 'zone', 'desc': 'The availability zone of host',
              'translator': value_trans})}

    floating_ips_translator = {
        'translation-type': 'HDICT',
        'table-name': FLOATING_IPS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'fixed_ip', 'desc': 'Fixed IP Address',
              'translator': value_trans},
             {'fieldname': 'id', 'desc': 'Unique ID',
              'translator': value_trans},
             {'fieldname': 'ip', 'desc': 'IP Address',
              'translator': value_trans},
             {'fieldname': 'instance_id',
              'desc': 'Name or ID of host', 'translator': value_trans},
             {'fieldname': 'pool', 'desc': 'Name of Floating IP Pool',
              'translator': value_trans})}

    services_translator = {
        'translation-type': 'HDICT',
        'table-name': SERVICES,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'col': 'service_id', 'desc': 'Service ID',
              'translator': value_trans},
             {'fieldname': 'binary', 'desc': 'Service binary',
              'translator': value_trans},
             {'fieldname': 'host', 'desc': 'Host Name',
              'translator': value_trans},
             {'fieldname': 'zone', 'desc': 'Availability Zone',
              'translator': value_trans},
             {'fieldname': 'status', 'desc': 'Status of service',
              'translator': value_trans},
             {'fieldname': 'state', 'desc': 'State of service',
              'translator': value_trans},
             {'fieldname': 'updated_at', 'desc': 'Last updated time',
              'translator': value_trans},
             {'fieldname': 'disabled_reason', 'desc': 'Disabled reason',
              'translator': value_trans})}

    availability_zones_translator = {
        'translation-type': 'HDICT',
        'table-name': AVAILABILITY_ZONES,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'zoneName', 'col': 'zone',
              'desc': 'Availability zone name', 'translator': value_trans},
             {'fieldname': 'zoneState', 'col': 'state',
              'desc': 'Availability zone state',
              'translator': value_trans})}

    TRANSLATORS = [servers_translator, flavors_translator, hosts_translator,
                   floating_ips_translator, services_translator,
                   availability_zones_translator]

    def __init__(self, name='', keys='', inbox=None, datapath=None, args=None):
        super(NovaDriver, self).__init__(name, keys, inbox, datapath, args)
        datasource_driver.ExecutionDriver.__init__(self)
        self.creds = self.get_nova_credentials_v2(args)
        self.nova_client = novaclient.client.Client(**self.creds)
        self.add_executable_method('servers_set_meta',
                                   [{'name': 'server',
                                    'description': 'server id'},
                                    {'name': 'meta',
                                     'description': 'metadata pairs, ' +
                                     'e.g. meta1=val1 meta2=val2'}],
                                   "A wrapper for servers.set_meta()")
        self.add_executable_client_methods(self.nova_client, 'novaclient.v2.')
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'nova'
        result['description'] = ('Datasource driver that interfaces with '
                                 'OpenStack Compute aka nova.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['config']['api_version'] = constants.OPTIONAL
        result['secret'] = ['password']
        return result

    def get_nova_credentials_v2(self, creds):
        d = {}
        d['version'] = creds.get('api_version', '2')
        d['username'] = creds['username']
        d['api_key'] = creds['password']
        d['auth_url'] = creds['auth_url']
        d['project_id'] = creds['tenant_name']
        return d

    def update_from_datasource(self):
        servers = self.nova_client.servers.list(
            detailed=True, search_opts={"all_tenants": 1})
        self._translate_servers(servers)
        self._translate_flavors(self.nova_client.flavors.list())
        self._translate_hosts(self.nova_client.hosts.list())
        self._translate_floating_ips(self.nova_client.floating_ips.list())
        self._translate_services(self.nova_client.services.list())
        self._translate_availability_zones(
            self.nova_client.availability_zones.list())

    @ds_utils.update_state_on_changed(SERVERS)
    def _translate_servers(self, obj):
        row_data = NovaDriver.convert_objs(obj, NovaDriver.servers_translator)
        return row_data

    @ds_utils.update_state_on_changed(FLAVORS)
    def _translate_flavors(self, obj):
        row_data = NovaDriver.convert_objs(obj, NovaDriver.flavors_translator)
        return row_data

    @ds_utils.update_state_on_changed(HOSTS)
    def _translate_hosts(self, obj):
        row_data = NovaDriver.convert_objs(obj, NovaDriver.hosts_translator)
        return row_data

    @ds_utils.update_state_on_changed(FLOATING_IPS)
    def _translate_floating_ips(self, obj):
        row_data = NovaDriver.convert_objs(obj,
                                           NovaDriver.floating_ips_translator)
        return row_data

    @ds_utils.update_state_on_changed(SERVICES)
    def _translate_services(self, obj):
        row_data = NovaDriver.convert_objs(obj, NovaDriver.services_translator)
        return row_data

    @ds_utils.update_state_on_changed(AVAILABILITY_ZONES)
    def _translate_availability_zones(self, obj):
        row_data = NovaDriver.convert_objs(
            obj,
            NovaDriver.availability_zones_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.nova_client, action, action_args)

    # "action" methods - to be used with "execute"
    def servers_set_meta(self, args):
        """A wrapper for servers.set_meta().

        'execute[p(x)]' doesn't take optional args at the moment.
        Therefore, this function translates the positional ARGS
        to optional args and call the servers.set_meta() api.
        :param <list> args: expected server ID and pairs of meta
        data in positional args such as:
        {'positional': ['server_id', 'meta1', 'value1', 'meta2', 'value2']}

        Usage:
        execute[nova.servers_set_meta(svr_id, meta1, val1, meta2, val2) :-
            triggering_table(id)
        """
        action = 'servers.set_meta'
        positional_args = args.get('positional', [])
        if not positional_args:
            LOG.error('Args not found for servers_set_meta()')
            return

        # Strip off the server_id before converting meta data pairs
        server_id = positional_args.pop(0)
        meta_data = self._convert_args(positional_args)

        action_args = {'named': {'server': server_id,
                                 'metadata': meta_data}}
        self._execute_api(self.nova_client, action, action_args)



# Copyright 2012 New Dream Network, LLC (DreamHost)
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os

from alembic import command as alembic_command
from alembic import config as alembic_config
from alembic import script as alembic_script
from alembic import util as alembic_util
from oslo_config import cfg

HEAD_FILENAME = 'HEAD'


_db_opts = [
    cfg.StrOpt('connection',
               deprecated_name='sql_connection',
               default='',
               secret=True,
               help=_('URL to database')),
    cfg.StrOpt('engine',
               default='',
               help=_('Database engine')),
]

CONF = cfg.ConfigOpts()
CONF.register_cli_opts(_db_opts, 'database')


def do_alembic_command(config, cmd, *args, **kwargs):
    try:
        getattr(alembic_command, cmd)(config, *args, **kwargs)
    except alembic_util.CommandError as e:
        alembic_util.err(str(e))


def do_check_migration(config, cmd):
    do_alembic_command(config, 'branches')
    validate_head_file(config)


def do_upgrade_downgrade(config, cmd):
    if not CONF.command.revision and not CONF.command.delta:
        raise SystemExit(_('You must provide a revision or relative delta'))

    revision = CONF.command.revision

    if CONF.command.delta:
        sign = '+' if CONF.command.name == 'upgrade' else '-'
        revision = sign + str(CONF.command.delta)
    else:
        revision = CONF.command.revision

    do_alembic_command(config, cmd, revision, sql=CONF.command.sql)


def do_stamp(config, cmd):
    do_alembic_command(config, cmd,
                       CONF.command.revision,
                       sql=CONF.command.sql)


def do_revision(config, cmd):
    do_alembic_command(config, cmd,
                       message=CONF.command.message,
                       autogenerate=CONF.command.autogenerate,
                       sql=CONF.command.sql)
    update_head_file(config)


def validate_head_file(config):
    script = alembic_script.ScriptDirectory.from_config(config)
    if len(script.get_heads()) > 1:
        alembic_util.err(_('Timeline branches unable to generate timeline'))

    head_path = os.path.join(script.versions, HEAD_FILENAME)
    if (os.path.isfile(head_path) and
            open(head_path).read().strip() == script.get_current_head()):
        return
    else:
        alembic_util.err(_('HEAD file does not match migration timeline head'))


def update_head_file(config):
    script = alembic_script.ScriptDirectory.from_config(config)
    if len(script.get_heads()) > 1:
        alembic_util.err(_('Timeline branches unable to generate timeline'))

    head_path = os.path.join(script.versions, HEAD_FILENAME)
    with open(head_path, 'w+') as f:
        f.write(script.get_current_head())


def add_command_parsers(subparsers):
    for name in ['current', 'history', 'branches']:
        parser = subparsers.add_parser(name)
        parser.set_defaults(func=do_alembic_command)

    parser = subparsers.add_parser('check_migration')
    parser.set_defaults(func=do_check_migration)

    for name in ['upgrade', 'downgrade']:
        parser = subparsers.add_parser(name)
        parser.add_argument('--delta', type=int)
        parser.add_argument('--sql', action='store_true')
        parser.add_argument('revision', nargs='?')
        parser.add_argument('--mysql-engine',
                            default='',
                            help='Change MySQL storage engine of current '
                                 'existing tables')
        parser.set_defaults(func=do_upgrade_downgrade)

    parser = subparsers.add_parser('stamp')
    parser.add_argument('--sql', action='store_true')
    parser.add_argument('revision')
    parser.set_defaults(func=do_stamp)

    parser = subparsers.add_parser('revision')
    parser.add_argument('-m', '--message')
    parser.add_argument('--autogenerate', action='store_true')
    parser.add_argument('--sql', action='store_true')
    parser.set_defaults(func=do_revision)


command_opt = cfg.SubCommandOpt('command',
                                title='Command',
                                help=_('Available commands'),
                                handler=add_command_parsers)

CONF.register_cli_opt(command_opt)


def main():
    config = alembic_config.Config(
        os.path.join(os.path.dirname(__file__), 'alembic.ini')
    )
    config.set_main_option('script_location',
                           'congress.db.migration:alembic_migrations')
    # attach the Neutron conf to the Alembic conf
    config.congress_config = CONF

    CONF()
    # TODO(gongysh) enable logging
    CONF.command.func(config, CONF.command.name)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Congress base exception handling."""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import sys

from oslo_config import cfg
from oslo_log import log as logging
import six

from congress.api import error_codes
from congress import utils


LOG = logging.getLogger(__name__)

exc_log_opts = [
    cfg.BoolOpt('fatal_exception_format_errors',
                default=False,
                help='Make exception message format errors fatal'),
]

CONF = cfg.CONF
CONF.register_opts(exc_log_opts)


class CongressException(Exception):
    """Base Congress Exception

    To correctly use this class, inherit from it and define
    a 'msg_fmt' property. That msg_fmt will get printf'd
    with the keyword arguments provided to the constructor.

    """
    msg_fmt = _("An unknown exception occurred.")
    # FIXME(thinrichs): this exception is overly complex and should
    #  not include HTTP codes at all.  Proper fix needs to touch
    #  too many files that others are currently working on.
    code = 500
    headers = {}
    safe = False

    def __init__(self, message=None, **kwargs):
        """FIXME(thinrichs):

        We just want name and data as fields.
        :param name will be a name from error_codes, which includes the basic
        message.
        :param data will contain specifics for this instance of the
        exception, e.g. a description error message.
        """
        self.data = kwargs.get('data', None)
        self.name = kwargs.get('name', None)

        # TODO(thinrichs): remove the rest of this (except the call to super)
        self.kwargs = kwargs
        if 'code' not in self.kwargs:
            try:
                self.kwargs['code'] = self.code
            except AttributeError:
                pass

        if not message:
            if self.name is not None:
                error_code = error_codes.get_num(self.name)
                description = error_codes.get_desc(self.name)
                message = "(%s) %s" % (error_code, description)
            else:
                try:
                    message = self.msg_fmt % kwargs

                except Exception:
                    exc_info = sys.exc_info()
                    # kwargs doesn't match a variable in the message
                    # log the issue and the kwargs
                    LOG.exception(_('Exception in string format operation'))
                    for name, value in kwargs.items():
                        LOG.error("%s: %s", name, value)    # noqa

                    if CONF.fatal_exception_format_errors:
                        six.reraise(exc_info[0], exc_info[1], exc_info[2])
                    else:
                        # at least get the core message out
                        message = self.msg_fmt

        super(CongressException, self).__init__(message)

    def format_message(self):
        # NOTE(mrodden): use the first argument to the python Exception object
        # which should be our full CongressException message, (see __init__)
        return self.args[0]

# FIXME(thinrichs): Get rid of the ones below and instead create exception
#   classes to represent the parts of the code that generated the exception,
#   e.g. datasources versus policy compiler versus policy runtime.


class Forbidden(CongressException):
    msg_fmt = _("Not authorized.")
    code = 403


class Conflict(CongressException):
    msg_fmt = _("Conflict")
    code = 409


class BadRequest(CongressException):
    msg_fmt = _("Bad request")
    code = 400


class NotFound(CongressException):
    msg_fmt = _("Resource not found.")
    code = 404


class PolicyNotAuthorized(Forbidden):
    msg_fmt = _("Policy doesn't allow %(action)s to be performed.")


class InvalidParamException(Exception):
    pass


class DataSourceConfigException(Exception):
    pass


class DuplicateTableName(Exception):
    pass


class InvalidTranslationType(Exception):
    pass


class DanglingReference(Conflict):
    pass


# NOTE(thinrichs): The following represent different kinds of
#   exceptions: the policy compiler and the policy runtime, respectively.
class PolicyException(CongressException):
    def __init__(self, msg=None, obj=None, line=None, col=None,
                 name=None, data=None, **kwargs):
        CongressException.__init__(self, message=msg, name=name, data=data)
        self.obj = obj
        self.location = utils.Location(line=line, col=col, obj=obj)

    def __str__(self):
        s = str(self.location)
        if len(s) > 0:
            s = " at" + s
        return CongressException.__str__(self) + s


class PolicyRuntimeException(CongressException):
    pass


class IncompleteSchemaException(CongressException):
    pass


class DataServiceError (Exception):
    pass


class BadConfig(BadRequest):
    pass


class DatasourceDriverException(CongressException):
    pass


class MissingRequiredConfigOptions(BadConfig):
    msg_fmt = _("Missing required config options: %(missing_options)s")


class InvalidDriver(BadConfig):
    msg_fmt = _("Invalid driver: %(driver)s")


class InvalidDriverOption(BadConfig):
    msg_fmt = _("Invalid driver options: %(invalid_options)s")


class DatasourceNameInUse(Conflict):
    msg_fmt = _("Datasource already in use with name %(value)s")


class DatasourceNotFound(NotFound):
    msg_fmt = _("Datasource not found %(id)s")


class DriverNotFound(NotFound):
    msg_fmt = _("Driver not found %(id)s")


class DatasourceCreationError(BadConfig):
    msg_fmt = _("Datasource could not be created on the DSE: %(value)s")



# Copyright (c) 2015 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_config import cfg
cfg.CONF.distributed_architecture = True

from congress.api import api_utils
from congress.api import schema_model
from congress.api import webservice
from congress.tests import base
from congress.tests2.api import base as api_base


class TestSchemaModel(base.TestCase):
    def setUp(self):
        super(TestSchemaModel, self).setUp()
        self.schema_model = schema_model.SchemaModel("test_schema", {})
        self.config = api_base.setup_config([self.schema_model])
        self.data = self.config['data']

    def test_get_item_all_table(self):
        context = {'ds_id': self.data.service_id}
        schema = self.data.get_schema()
        fake_tables = {'tables':
                       [api_utils.create_table_dict(
                        table_, schema) for table_ in schema]}
        tables = self.schema_model.get_item(None, {}, context=context)
        self.assertEqual(fake_tables, tables)

    def test_get_item_table(self):
        context = {'ds_id': self.data.service_id, 'table_id': 'fake_table'}
        fake_schema = self.data.get_schema()
        fake_table = api_utils.create_table_dict(
            "fake_table", fake_schema)
        table = self.schema_model.get_item(None, {}, context=context)
        self.assertEqual(fake_table, table)

    def test_get_invalid_datasource_table(self):
        context = {'ds_id': self.data.service_id, 'table_id': 'invalid_table'}
        try:
            self.schema_model.get_item(None, {}, context=context)
        except webservice.DataModelException as e:
            self.assertEqual(404, e.error_code)
        else:
            raise Exception("Should not get here")

    def test_get_invalid_datasource(self):
        context = {'ds_id': 'invalid'}
        try:
            self.schema_model.get_item(None, {}, context=context)
        except webservice.DataModelException as e:
            self.assertEqual(404, e.error_code)
        else:
            raise Exception("Should not get here")



# Copyright (c) 2015 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

# TODO(thinrichs): move algorithms from compile.py that do analysis
# into this file.

import copy


class ModalIndex(object):
    def __init__(self):
        # Dict mapping modal name to a ref-counted list of tablenames
        # Refcounted list of tablenames is a dict from tablename to count
        self.index = {}

    def add(self, modal, tablename):
        if modal not in self.index:
            self.index[modal] = {}
        if tablename not in self.index[modal]:
            self.index[modal][tablename] = 0
        self.index[modal][tablename] += 1

    def remove(self, modal, tablename):
        if modal not in self.index:
            raise KeyError("Modal %s has no entries" % modal)
        if tablename not in self.index[modal]:
            raise KeyError("Tablename %s for modal %s does not exist" %
                           (tablename, modal))
        self.index[modal][tablename] -= 1
        self._clean_up(modal, tablename)

    def modals(self):
        return self.index.keys()

    def tables(self, modal):
        if modal not in self.index:
            return []
        return self.index[modal].keys()

    def __isub__(self, other):
        changes = []
        for modal in self.index:
            if modal not in other.index:
                continue
            for table in self.index[modal]:
                if table not in other.index[modal]:
                    continue
                self.index[modal][table] -= other.index[modal][table]
                changes.append((modal, table))

        for (modal, table) in changes:
            self._clean_up(modal, table)
        return self

    def __iadd__(self, other):
        for modal in other.index:
            if modal not in self.index:
                self.index[modal] = other.index[modal]
                continue
            for table in other.index[modal]:
                if table not in self.index[modal]:
                    self.index[modal][table] = other.index[modal][table]
                    continue
                self.index[modal][table] += other.index[modal][table]
        return self

    def _clean_up(self, modal, table):
        if self.index[modal][table] <= 0:
            del self.index[modal][table]
        if not len(self.index[modal]):
            del self.index[modal]

    def __eq__(self, other):
        return self.index == other.index

    def __neq__(self, other):
        return not self.__eq__(other)

    def __copy__(self):
        new = ModalIndex()
        new.index = copy.deepcopy(self.index)
        return new

    def __str__(self):
        return str(self.index)

    def __contains__(self, modal):
        return modal in self.index



# Copyright 2014 VMware.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from django.conf.urls import patterns
from django.conf.urls import url
from openstack_dashboard.dashboards.admin.datasources import views


SERVICES = (
    r'^services/(?P<datasource_id>[^/]+)/(?P<service_table_name>[^/]+)/%s$')
POLICIES = (
    r'^policies/(?P<datasource_id>[^/]+)/(?P<policy_table_name>[^/]+)/%s$')


urlpatterns = patterns(
    '',
    url(r'^$', views.IndexView.as_view(), name='index'),
    url(SERVICES % 'detail', views.DetailView.as_view(),
        name='datasource_table_detail'),
    url(POLICIES % 'detail', views.DetailView.as_view(),
        name='policy_table_detail'),
)



# Copyright (c) 2016 NEC Corporation. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

""" Base class for all API models."""

from __future__ import absolute_import

from oslo_config import cfg

# Use new deepsix when appropriate
if getattr(cfg.CONF, 'distributed_architecture', None):
    from congress.dse2 import deepsix2 as deepsix
else:
    from congress.dse import deepsix
from congress import exception


class APIModel(deepsix.deepSix):
    """Base Class for handling API requests."""

    def __init__(self, name, keys='', inbox=None, dataPath=None,
                 policy_engine=None, datasource_mgr=None):
        super(APIModel, self).__init__(name, keys, inbox=inbox,
                                       dataPath=dataPath)
        self.engine = policy_engine
        self.datasource_mgr = datasource_mgr

    def invoke_rpc(self, caller, name, kwargs):
        if getattr(cfg.CONF, 'distributed_architecture', None):
            return self.rpc(caller, name, kwargs)
        else:
            func = getattr(caller, name, None)
            if func:
                return func(**kwargs)
            raise exception.CongressException('method: %s is not defined in %s'
                                              % (name, caller.__name__))



# Copyright (c) 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import json

from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import uuidutils
import six

from congress.datalog import base
from congress.datasources import constants
from congress.db import api as db
from congress.db import datasources as datasources_db
from congress.dse import d6cage
from congress import exception

import sys
import traceback


LOG = logging.getLogger(__name__)


class DataServiceError (Exception):
    pass


class DataSourceManager(object):

    loaded_drivers = {}
    dseNode = None

    @classmethod
    def set_dseNode(cls, dseNode):
        cls.dseNode = dseNode

    @classmethod
    def add_datasource(cls, item, deleted=False, update_db=True):
        req = cls.make_datasource_dict(item)
        # If update_db is True, new_id will get a new value from the db.
        new_id = req['id']
        driver_info = cls.get_driver_info(item['driver'])
        session = db.get_session()
        try:
            with session.begin(subtransactions=True):
                LOG.debug("adding datasource %s", req['name'])
                if update_db:
                    LOG.debug("updating db")
                    datasource = datasources_db.add_datasource(
                        id_=req['id'],
                        name=req['name'],
                        driver=req['driver'],
                        config=req['config'],
                        description=req['description'],
                        enabled=req['enabled'],
                        session=session)
                    new_id = datasource['id']

                cls.validate_create_datasource(req)
                cage = cls.dseNode or d6cage.d6Cage()
                engine = cage.service_object('engine')
                try:
                    LOG.debug("creating policy %s", req['name'])
                    engine.create_policy(req['name'],
                                         kind=base.DATASOURCE_POLICY_TYPE)
                except KeyError:
                    # FIXME(arosen): we need a better exception then
                    # key error being raised here
                    raise exception.DatasourceNameInUse(value=req['name'])
                try:
                    if cls.dseNode:
                        cls.createservice(name=req['name'],
                                          moduleName=driver_info['module'],
                                          args=item['config'],
                                          module_driver=True,
                                          type_='datasource_driver',
                                          id_=new_id)
                    else:
                        cage.createservice(name=req['name'],
                                           moduleName=driver_info['module'],
                                           args=item['config'],
                                           module_driver=True,
                                           type_='datasource_driver',
                                           id_=new_id)
                    service = cage.service_object(req['name'])
                    engine.set_schema(req['name'], service.get_schema())
                except Exception:
                    engine.delete_policy(req['name'])
                    raise exception.DatasourceCreationError(value=req['name'])

        except db_exc.DBDuplicateEntry:
            raise exception.DatasourceNameInUse(value=req['name'])
        new_item = dict(item)
        new_item['id'] = new_id
        return cls.make_datasource_dict(new_item)

    @classmethod
    def validate_configured_drivers(cls):
        """load all configured drivers and check no name conflict"""
        result = {}
        for driver_path in cfg.CONF.drivers:
            obj = importutils.import_class(driver_path)
            driver = obj.get_datasource_info()
            if driver['id'] in result:
                raise exception.BadConfig(_("There is a driver loaded already"
                                          "with the driver name of %s") %
                                          driver['id'])
            driver['module'] = driver_path
            result[driver['id']] = driver
        cls.loaded_drivers = result

    @classmethod
    def make_datasource_dict(cls, req, fields=None):
        result = {'id': req.get('id') or uuidutils.generate_uuid(),
                  'name': req.get('name'),
                  'driver': req.get('driver'),
                  'description': req.get('description'),
                  'type': None,
                  'enabled': req.get('enabled', True)}
        # NOTE(arosen): we store the config as a string in the db so
        # here we serialize it back when returning it.
        if isinstance(req.get('config'), six.string_types):
            result['config'] = json.loads(req['config'])
        else:
            result['config'] = req.get('config')

        return cls._fields(result, fields)

    @classmethod
    def _fields(cls, resource, fields):
        if fields:
            return dict(((key, item) for key, item in resource.items()
                         if key in fields))
        return resource

    @classmethod
    def get_datasources(cls, filter_secret=False):
        """Return the created datasources.

        This returns what datasources the database contains, not the
        datasources that this server instance is running.
        """

        results = []
        for datasouce_driver in datasources_db.get_datasources():
            result = cls.make_datasource_dict(datasouce_driver)
            if filter_secret:
                # secret field may be not provided while creating datasource
                try:
                    hides = cls.get_driver_info(result['driver'])['secret']
                    for hide_field in hides:
                        result['config'][hide_field] = "<hidden>"
                except KeyError:
                    pass
            results.append(result)
        return results

    @classmethod
    def get_datasource(cls, id_):
        """Return the created datasource."""
        result = datasources_db.get_datasource(id_)
        if not result:
            raise exception.DatasourceNotFound(id=id_)
        return cls.make_datasource_dict(result)

    @classmethod
    def get_driver_info(cls, driver):
        driver = cls.loaded_drivers.get(driver)
        if not driver:
            raise exception.DriverNotFound(id=driver)
        return driver

    @classmethod
    def get_driver_schema(cls, datasource_id):
        driver = cls.get_driver_info(datasource_id)
        obj = importutils.import_class(driver['module'])
        return obj.get_schema()

    @classmethod
    def get_datasource_schema(cls, source_id):
        datasource = datasources_db.get_datasource(source_id)
        if not datasource:
            raise exception.DatasourceNotFound(id=source_id)
        driver = cls.get_driver_info(datasource.driver)
        if driver:
            # NOTE(arosen): raises if not found
            driver = cls.get_driver_info(
                driver['id'])
            obj = importutils.import_class(driver['module'])
            return obj.get_schema()

    @classmethod
    def load_module_object(cls, datasource_id_or_name):
        datasource = datasources_db.get_datasource(datasource_id_or_name)
        # Ideally speaking, it should change datasource_db.get_datasource() to
        # be able to retrieve datasource info from db at once. The datasource
        # table and the method, however, will be removed in the new
        # architecture, so it use this way. Supporting both name and id is
        # a backward compatibility.
        if not datasource:
            datasource = (datasources_db.
                          get_datasource_by_name(datasource_id_or_name))
        if not datasource:
            return None

        driver = cls.get_driver_info(datasource.driver)
        obj = importutils.import_class(driver['module'])

        return obj

    @classmethod
    def get_row_data(cls, table_id, source_id, **kwargs):
        datasource = cls.get_datasource(source_id)
        cage = cls.dseNode or d6cage.d6Cage()
        datasource_obj = cage.service_object(datasource['name'])
        return datasource_obj.get_row_data(table_id)

    @classmethod
    def update_entire_data(cls, table_id, source_id, objs):
        datasource = cls.get_datasource(source_id)
        cage = d6cage.d6Cage()
        datasource_obj = cage.service_object(datasource['name'])
        return datasource_obj.update_entire_data(table_id, objs)

    @classmethod
    def get_tablename(cls, source_id, table_id):
        obj = cls.load_module_object(source_id)
        if obj:
            return obj.get_tablename(table_id)
        else:
            return None

    @classmethod
    def get_tablenames(cls, source_id):
        '''The method to get datasource tablename.'''
        # In the new architecture, table model would call datasource_driver's
        # get_tablenames() directly using RPC
        obj = cls.load_module_object(source_id)

        if obj:
            return obj.get_tablenames()
        else:
            return None

    @classmethod
    def delete_datasource(cls, datasource_id, update_db=True):
        datasource = cls.get_datasource(datasource_id)
        session = db.get_session()
        with session.begin(subtransactions=True):
            cage = cls.dseNode or d6cage.d6Cage()
            engine = cage.service_object('engine')
            try:
                engine.delete_policy(datasource['name'],
                                     disallow_dangling_refs=True)
            except exception.DanglingReference as e:
                raise e
            except KeyError:
                raise exception.DatasourceNotFound(id=datasource_id)
            if update_db:
                result = datasources_db.delete_datasource(
                    datasource_id, session)
                if not result:
                    raise exception.DatasourceNotFound(id=datasource_id)
            if cls.dseNode:
                cls.dseNode.unregister_service(
                    cls.dseNode.service_object(datasource['name']))
            else:
                cage.deleteservice(datasource['name'])

    @classmethod
    def get_status(cls, source_id=None, params=None):
        cage = d6cage.d6Cage()
        driver = cage.getservice(id_=source_id, type_='datasource_driver')
        if not driver:
            raise exception.NotFound('Could not find datasource %s' %
                                     source_id)
        return driver['object'].get_status()

    @classmethod
    def get_actions(cls, source_id=None):
        cage = d6cage.d6Cage()
        driver = cage.getservice(id_=source_id, type_='datasource_driver')
        if not driver:
            raise exception.NotFound('Could not find datasource %s' %
                                     source_id)
        return driver['object'].get_actions()

    @classmethod
    def get_drivers_info(cls):
        return [driver for driver in cls.loaded_drivers.values()]

    @classmethod
    def validate_create_datasource(cls, req):
        driver = req['driver']
        config = req['config'] or {}
        for loaded_driver in cls.loaded_drivers.values():
            if loaded_driver['id'] == driver:
                specified_options = set(config.keys())
                valid_options = set(loaded_driver['config'].keys())
                # Check that all the specified options passed in are
                # valid configuration options that the driver exposes.
                invalid_options = specified_options - valid_options
                if invalid_options:
                    raise exception.InvalidDriverOption(
                        invalid_options=invalid_options)

                # check that all the required options are passed in
                required_options = set(
                    [k for k, v in loaded_driver['config'].items()
                     if v == constants.REQUIRED])
                missing_options = required_options - specified_options
                if missing_options:
                    missing_options = ', '.join(missing_options)
                    raise exception.MissingRequiredConfigOptions(
                        missing_options=missing_options)
                return loaded_driver

        # If we get here no datasource driver match was found.
        raise exception.InvalidDriver(driver=req)

    @classmethod
    def request_refresh(cls, source_id):
        datasource = cls.get_datasource(source_id)
        cage = cls.dseNode or d6cage.d6Cage()
        datasource = cage.service_object(datasource['name'])
        datasource.request_refresh()

    @classmethod
    def createservice(
            cls,
            name="",
            keys="",
            description="",
            moduleName="",
            args={},
            module_driver=False,
            type_=None,
            id_=None):
        # copied from d6cage. It's not clear where this code should reside LT

        # self.log_info("creating service %s with module %s and args %s",
        #               name, moduleName, strutils.mask_password(args, "****"))

        # FIXME(arosen) This will be refactored out in the next patchset
        # this is only done because existing imports from d6service
        # instead of the module.
        if module_driver:
            congress_expected_module_path = ""
            for entry in range(len(moduleName.split(".")) - 1):
                congress_expected_module_path += (
                    moduleName.split(".")[entry] + ".")
            congress_expected_module_path = congress_expected_module_path[:-1]
            module = importutils.import_module(congress_expected_module_path)

        if not module_driver and moduleName not in sys.modules:
            # self.log_error(
            #     "error loading service %s: module %s does not exist",
            #     name,
            #     moduleName)
            raise exception.DataServiceError(
                "error loading service %s: module %s does not exist" %
                (name, moduleName))

        # if not module_driver and name in self.services:
        #     self.log_error("error loading service '%s': name already in use",
        #                         name)
        #     raise DataServiceError(
        #         "error loading service '%s': name already in use"
        #         % name)

        if not module_driver:
            module = sys.modules[moduleName]

        try:
            svcObject = module.d6service(name, keys, None, None,
                                         args)
            cls.dseNode.register_service(svcObject)
        except Exception:
            # self.log_error(
            #            "Error loading service '%s' of module '%s':: \n%s",
            #            name, module, traceback.format_exc())
            raise exception.DataServiceError(
                "Error loading service '%s' of module '%s':: \n%s"
                % (name, module, traceback.format_exc()))



# Copyright (c) 2013 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import copy

from congress.datalog import analysis
from congress.datalog import base as datalogbase
from congress.datalog import compile
from congress.datalog import utility
from congress import exception
from congress.policy_engines import agnostic
from congress.tests import base
from congress.tests import helper


class TestParser(base.TestCase):

    def test_tablename(self):
        """Test correct parsing of tablenames."""
        p = compile.parse1('p(1)')
        self.assertEqual(p.table.table, 'p')
        self.assertIsNone(p.table.modal)
        self.assertIsNone(p.table.service)

        p = compile.parse1('nova:p(1)')
        self.assertEqual(p.table.table, 'p')
        self.assertIsNone(p.table.modal)
        self.assertEqual(p.table.service, 'nova')

        p = compile.parse1('execute[nova:p(1)]')
        self.assertEqual(p.table.table, 'p')
        self.assertEqual(p.table.modal, 'execute')
        self.assertEqual(p.table.service, 'nova')

    def test_rule_hash(self):
        """Test whether rules are properly hashed."""
        s = set()
        p = compile.parse1('p(x) :- q(x), s(x,y)')
        s.add(p)
        q = compile.parse1('p(x) :- q(x), s(x,y)')
        s.discard(q)
        self.assertEqual(s, set())

    def test_event_equality(self):
        r1 = compile.parse1('p(x) :- q(x)')
        r2 = compile.parse1('p(x) :- q(x)')
        e1 = compile.Event(formula=r1, target='alice', insert=True)
        e2 = compile.Event(formula=r2, target='alice', insert=True)
        self.assertEqual(e1, e2)

    def test_event_facts(self):
        # insert
        event = compile.parse('insert[p(1) :- true]')
        self.assertEqual(len(event), 1)
        event = event[0]
        fact = compile.parse1('p(1) :- true')
        self.assertEqual(event.formula, fact)
        self.assertEqual(event.insert, True)
        self.assertIsNone(event.target)

        # delete
        event = compile.parse('delete[p(1) :- true]')
        self.assertEqual(len(event), 1)
        event = event[0]
        fact = compile.parse1('p(1) :- true')
        self.assertEqual(event.formula, fact)
        self.assertEqual(event.insert, False)
        self.assertIsNone(event.target)

        # insert with policy
        event = compile.parse('insert[p(1) :- true; "policy"]')
        self.assertEqual(len(event), 1)
        event = event[0]
        fact = compile.parse1('p(1) :- true')
        self.assertEqual(event.formula, fact)
        self.assertEqual(event.insert, True)
        self.assertEqual(event.target, "policy")

    def test_event_rules(self):
        """Test modal operators."""
        # a rule we use a few times
        pqrule = compile.parse1('p(x) :- q(x)')

        # rule-level modal (with insert)
        event = compile.parse('insert[p(x) :- q(x)]')
        self.assertEqual(len(event), 1)
        event = event[0]
        self.assertEqual(event.formula, pqrule)
        self.assertEqual(event.insert, True)
        self.assertIsNone(event.target)

        # rule-level modal with delete
        event = compile.parse('delete[p(x) :- q(x)]')
        self.assertEqual(len(event), 1)
        event = event[0]
        self.assertEqual(event.formula, pqrule)
        self.assertEqual(event.insert, False)
        self.assertIsNone(event.target)

        # embedded modals
        event = compile.parse('insert[execute[p(x)] :- q(x)]')
        self.assertEqual(len(event), 1)
        event = event[0]
        rule = compile.parse1('execute[p(x)] :- q(x)')
        self.assertEqual(event.formula, rule)
        self.assertEqual(event.insert, True)
        self.assertIsNone(event.target)

        # rule-level modal with policy name
        event = compile.parse('insert[p(x) :- q(x); "policy"]')
        self.assertEqual(len(event), 1)
        event = event[0]
        self.assertEqual(event.formula, pqrule)
        self.assertEqual(event.insert, True)
        self.assertEqual(event.target, "policy")

    def test_modal_execute(self):
        # modal rule
        rule = compile.parse('execute[p(x)] :- q(x)')
        self.assertEqual(len(rule), 1)
        rule = rule[0]
        self.assertEqual(rule.head.table.modal, 'execute')

        # modal rule with namespace
        rule = compile.parse('execute[nova:disconnectNetwork(x)] :- q(x)')
        self.assertEqual(len(rule), 1)
        rule = rule[0]
        self.assertEqual(rule.head.table.modal, 'execute')

        # modal query
        rule = compile.parse('execute[p(x)]')
        self.assertEqual(len(rule), 1)
        rule = rule[0]
        self.assertEqual(rule.table.modal, 'execute')

    def test_update_rules(self):
        rule = compile.parse1('insert[p(x)] :- q(x)')
        self.assertEqual(rule.head.table.modal, 'insert')

        rule = compile.parse1('insert[p(x)] :- execute[q(x)]')
        self.assertEqual(rule.head.table.modal, 'insert')

    def test_modal_failures(self):
        self.assertRaises(exception.PolicyException, compile.parse1,
                          'insert[p(x) :- q(x)')
        self.assertRaises(exception.PolicyException, compile.parse1,
                          'insert[insert[p(x)] :- q(x)')
        self.assertRaises(exception.PolicyException, compile.parse1,
                          'nonexistent[insert[p(x)] :- q(x)]')
        self.assertRaises(exception.PolicyException, compile.parse1,
                          'insert[nonexistent[p(x)] :- q(x)]')


class TestColumnReferences(base.TestCase):

    class SchemaWrapper (object):
        """Placeholder so we don't use the actual policy-engine for tests."""
        def __init__(self, schema):
            self.schema = schema
            self.kind = datalogbase.DATASOURCE_POLICY_TYPE

    def test_column_references_lowlevel(self):
        """Test column-references with low-level checks."""
        # do the first one the painful way, to ensure the parser
        #   is doing something reasonable.
        run = agnostic.Runtime()
        code = ("p(x) :- nova:q(id=x)")
        actual = run.parse(code)
        self.assertEqual(len(actual), 1)
        rule = actual[0]
        self.assertEqual(len(rule.heads), 1)
        self.assertEqual(rule.head.table.table, "p")
        self.assertEqual(len(rule.head.arguments), 1)
        self.assertEqual(rule.head.arguments[0].name, 'x')
        self.assertEqual(len(rule.body), 1)
        lit = rule.body[0]
        self.assertFalse(lit.is_negated())
        self.assertEqual(lit.table.table, "q")
        self.assertIsNone(lit.table.modal)
        self.assertEqual(lit.table.service, 'nova')
        self.assertEqual(len(lit.arguments), 0)
        self.assertEqual(len(lit.named_arguments), 1)
        self.assertTrue('id' in lit.named_arguments)
        self.assertEqual(lit.named_arguments['id'].name, 'x')

    def test_hash(self):
        x = set()
        x.add(compile.parse1('p(x, y) :- nova:q(x, id=y)'))
        x.add(compile.parse1('p(x, y) :- nova:q(x, id=y)'))
        self.assertEqual(len(x), 1)
        self.assertTrue(compile.parse1('p(x, y) :- nova:q(x, id=y)') in x)
        x.discard(compile.parse1('p(x, y) :- nova:q(x, id=y)'))
        self.assertEqual(len(x), 0)

    def test_lessthan(self):
        x = compile.parse1('nova:q(x)')
        y = compile.parse1('nova:q(x, id=y)')
        self.assertLess(x, y)

        x = compile.parse1('nova:q(x)')
        y = compile.parse1('nova:q(x, id=y)')
        self.assertGreaterEqual(y, x)

        x = compile.parse1('nova:q(x, id=w)')
        y = compile.parse1('nova:q(x, id=y)')
        self.assertLess(x, y)

        x = compile.parse1('nova:q(id=x)')
        y = compile.parse1('nova:q(id=y)')
        self.assertLess(x, y)

        x = compile.parse1('nova:q(id=x)')
        y = compile.parse1('nova:q(id=y, status=z)')
        self.assertLess(x, y)

        x = compile.parse1('p(x, y) :- nova:q(x, id=y)')
        y = compile.parse1('p(x, y) :- nova:q(x, id=y, status=z)')
        self.assertLess(x, y)

    def test_column_references_parse_errors(self):
        """Test invalid column references occurring in a single atom."""
        def check_err(code, errmsg, msg):
            try:
                compile.parse(code)
                self.fail("Error should have been thrown but was not: " + msg)
            except exception.PolicyException as e:
                emsg = "Err message '{}' should include '{}'".format(
                    str(e), errmsg)
                self.assertTrue(errmsg in str(e), msg + ": " + emsg)

        check_err(
            'p(x) :- nova:q(id=x, status=x, id=y)',
            'two values for column name id',
            'Multiple values for column name')

        check_err(
            'p(x) :- nova:q(4=y, id=w, 4=z)',
            'two values for column number 4',
            'Multiple values for column number')

        check_err(
            'p(x) :- nova:q(x, 1=z, y)',
            'positional parameter after a reference parameter',
            'Positional parameter after reference parameter')

        check_err(
            'p(x) :- nova:q(x, y, 1=z)',
            '1 is already provided by position arguments',
            'Conflict between name and position')

    def test_column_references_validation_errors(self):
        """Test invalid column references occurring in a single atom."""
        schema = compile.Schema({'q': ('id', 'name', 'status'),
                                 'r': ('id', 'age', 'weight')},
                                complete=True)
        theories = {'nova': self.SchemaWrapper(schema)}

        def check_err(rule, errmsg, msg):
            rule = compile.parse1(rule)
            try:
                rule.eliminate_column_references(theories)
                self.fail("Failed to throw error {}".format(errmsg))
            except (exception.PolicyException,
                    exception.IncompleteSchemaException) as e:
                emsg = "Err messages '{}' should include '{}'".format(
                    str(e), errmsg)
                self.assertTrue(errmsg in str(e), msg + ": " + emsg)

        check_err(
            'p(x) :- nova:missing(id=x)',
            'uses unknown table missing',
            'Unknown table')

        check_err(
            'p(x) :- nova:q(id=x, birthday=y)',
            'column name birthday does not exist',
            'Unknown column name')

        check_err(
            'p(x) :- nova:q(4=y)',
            'column index 4 is too large',
            'Large column number')

        check_err(
            'p(x) :- nova:q(id=x, 0=y)',
            'index 0 references column id, which is also referenced by name',
            'Conflict between name and number references')

        check_err(
            'p(x) :- nova:q(x, y, id=z)',
            'already provided by position',
            'Conflict between name and position')

        theories = {}
        check_err(
            'p(x) :- nova:missing(id=x)',
            'schema is unknown',
            'Missing schema')

    def test_column_references_atom(self):
        """Test column references occurring in a single atom in a rule."""
        def check(code, correct, msg):
            actual = compile.parse1(code).eliminate_column_references(theories)
            eq = helper.datalog_same(str(actual), correct)
            self.assertTrue(eq, msg)

        run = agnostic.Runtime()
        run.create_policy('nova')
        schema = compile.Schema({'q': ('id', 'name', 'status')})
        theories = {'nova': self.SchemaWrapper(schema)}

        # Multiple column names
        code = ("p(x) :- nova:q(id=x, status=y)")
        correct = "p(x) :- nova:q(x, w, y)"
        check(code, correct, 'Multiple column names')

        # Multiple column numbers
        code = ("p(x) :- nova:q(0=x, 1=y, 2=z)")
        correct = "p(x) :- nova:q(x, y, z)"
        check(code, correct, 'Multiple column numbers')

        # Mix column names and numbers
        code = ("p(x) :- nova:q(id=x, 2=y)")
        correct = "p(x) :- nova:q(x, w, y)"
        check(code, correct, 'Mix names and numbers')

        # Object constants
        code = ("p(x) :- nova:q(id=3, 2=2)")
        correct = "p(x) :- nova:q(3, w, 2)"
        check(code, correct, 'Object constants')

        # Out of order
        code = ("p(x, y) :- nova:q(status=y, id=x)")
        correct = "p(x, y) :- nova:q(x, z, y)"
        check(code, correct,  'Out of order')

        # Out of order with numbers
        code = ("p(x, y) :- nova:q(1=y, 0=x)")
        correct = "p(x, y) :- nova:q(x, y, z)"
        check(code, correct, 'Out of order with numbers')

        # Positional plus named
        code = ("p(x, y) :- nova:q(x, status=y)")
        correct = "p(x, y) :- nova:q(x, z, y)"
        check(code, correct, 'Positional plus named')

        # Positional plus named 2
        code = ("p(x, y, z) :- nova:q(x, y, 2=z)")
        correct = "p(x, y, z) :- nova:q(x, y, z)"
        check(code, correct, 'Positional plus named 2')

        # Pure positional (different since we are providing schema)
        code = ("p(x, y, z) :- nova:q(x, y, z)")
        correct = "p(x, y, z) :- nova:q(x, y, z)"
        check(code, correct, 'Pure positional')

        # Pure positional (without schema)
        code = ("p(x) :- nova:q(x, y, z)")
        run.delete_policy('nova')
        correct = "p(x) :- nova:q(x, y, z)"
        check(code, correct, 'Pure positional without schema')

    def test_column_references_multiple_atoms(self):
        """Test column references occurring in multiple atoms in a rule."""
        def check(code, correct, msg):
            actual = compile.parse1(code).eliminate_column_references(theories)
            eq = helper.datalog_same(str(actual), correct)
            self.assertTrue(eq, msg)

        run = agnostic.Runtime()
        run.create_policy('nova')
        schema = compile.Schema({'q': ('id', 'name', 'status'),
                                 'r': ('id', 'age', 'weight')})
        theories = {'nova': self.SchemaWrapper(schema)}

        # Multiple atoms
        code = ("p(x) :- nova:q(id=x, 2=y), nova:r(id=x)")
        correct = "p(x) :- nova:q(x, x0, y), nova:r(x, y0, y1)"
        check(code, correct, 'Multiple atoms')

        # Multiple atoms sharing column name but different variables
        code = ("p(x) :- nova:q(id=x), nova:r(id=y)")
        correct = "p(x) :- nova:q(x, x0, x1), nova:r(y, y0, y1)"
        check(code, correct, 'Multiple atoms shared column name')

        # Multiple atoms, same table
        code = ("p(x) :- nova:q(id=x, 2=y), nova:q(id=x)")
        correct = "p(x) :- nova:q(x, x0, y), nova:q(x, y0, y1)"
        check(code, correct, 'Multiple atoms, same table')

    def test_eliminate_column_references_body_order(self):
        """Test eliminate_column_references preserves order insensitivity."""
        run = agnostic.Runtime()
        run.create_policy('nova')
        schema = compile.Schema({'q': ('id', 'name', 'status'),
                                 'r': ('id', 'age', 'weight')})
        theories = {'nova': self.SchemaWrapper(schema)}

        rule1 = compile.parse1("p(x) :- nova:q(id=x, 2=y), nova:r(id=x)"
                               ).eliminate_column_references(theories)
        rule2 = compile.parse1("p(x) :- nova:r(id=x), nova:q(id=x, 2=y)"
                               ).eliminate_column_references(theories)
        self.assertEqual(rule1, rule2, 'eliminate_column_references failed to '
                                       'preserve order insensitivity')


class TestCompiler(base.TestCase):

    def test_type_checkers(self):
        """Test the type checkers, e.g. is_atom, is_rule."""
        atom = compile.Literal("p", [])
        atom2 = compile.Literal("q", [])
        atom3 = compile.Literal("r", [])
        lit = compile.Literal("r", [], negated=True)
        regular_rule = compile.Rule(atom, [atom2, atom3])
        regular_rule2 = compile.Rule(atom, [lit, atom2])
        multi_rule = compile.Rule([atom, atom2], [atom3])
        fake_rule = compile.Rule([atom, 1], [atom2])
        fake_rule2 = compile.Rule(atom, [atom2, 1])

        # is_atom
        self.assertTrue(compile.is_atom(atom))
        self.assertTrue(compile.is_atom(atom2))
        self.assertTrue(compile.is_atom(atom3))
        self.assertFalse(compile.is_atom(lit))
        self.assertFalse(compile.is_atom(regular_rule))
        self.assertFalse(compile.is_atom(regular_rule2))
        self.assertFalse(compile.is_atom(multi_rule))
        self.assertFalse(compile.is_atom(fake_rule))
        self.assertFalse(compile.is_atom(fake_rule2))
        self.assertFalse(compile.is_atom("a string"))

        # is_literal
        self.assertTrue(compile.is_literal(atom))
        self.assertTrue(compile.is_literal(atom2))
        self.assertTrue(compile.is_literal(atom3))
        self.assertTrue(compile.is_literal(lit))
        self.assertFalse(compile.is_literal(regular_rule))
        self.assertFalse(compile.is_literal(regular_rule2))
        self.assertFalse(compile.is_literal(multi_rule))
        self.assertFalse(compile.is_literal(fake_rule))
        self.assertFalse(compile.is_literal(fake_rule2))
        self.assertFalse(compile.is_literal("a string"))

        # is_regular_rule
        self.assertFalse(compile.is_regular_rule(atom))
        self.assertFalse(compile.is_regular_rule(atom2))
        self.assertFalse(compile.is_regular_rule(atom3))
        self.assertFalse(compile.is_regular_rule(lit))
        self.assertTrue(compile.is_regular_rule(regular_rule))
        self.assertTrue(compile.is_regular_rule(regular_rule2))
        self.assertFalse(compile.is_regular_rule(multi_rule))
        self.assertFalse(compile.is_regular_rule(fake_rule))
        self.assertFalse(compile.is_regular_rule(fake_rule2))
        self.assertFalse(compile.is_regular_rule("a string"))

        # is_multi_rule
        self.assertFalse(compile.is_multi_rule(atom))
        self.assertFalse(compile.is_multi_rule(atom2))
        self.assertFalse(compile.is_multi_rule(atom3))
        self.assertFalse(compile.is_multi_rule(lit))
        self.assertFalse(compile.is_multi_rule(regular_rule))
        self.assertFalse(compile.is_multi_rule(regular_rule2))
        self.assertTrue(compile.is_multi_rule(multi_rule))
        self.assertFalse(compile.is_multi_rule(fake_rule))
        self.assertFalse(compile.is_multi_rule(fake_rule2))
        self.assertFalse(compile.is_multi_rule("a string"))

        # is_rule
        self.assertFalse(compile.is_rule(atom))
        self.assertFalse(compile.is_rule(atom2))
        self.assertFalse(compile.is_rule(atom3))
        self.assertFalse(compile.is_rule(lit))
        self.assertTrue(compile.is_rule(regular_rule))
        self.assertTrue(compile.is_rule(regular_rule2))
        self.assertTrue(compile.is_rule(multi_rule))
        self.assertFalse(compile.is_rule(fake_rule))
        self.assertFalse(compile.is_rule(fake_rule2))
        self.assertFalse(compile.is_rule("a string"))

        # is_datalog
        self.assertTrue(compile.is_datalog(atom))
        self.assertTrue(compile.is_datalog(atom2))
        self.assertTrue(compile.is_datalog(atom3))
        self.assertFalse(compile.is_datalog(lit))
        self.assertTrue(compile.is_datalog(regular_rule))
        self.assertTrue(compile.is_datalog(regular_rule2))
        self.assertFalse(compile.is_datalog(multi_rule))
        self.assertFalse(compile.is_datalog(fake_rule))
        self.assertFalse(compile.is_datalog(fake_rule2))
        self.assertFalse(compile.is_datalog("a string"))

        # is_extended_datalog
        self.assertTrue(compile.is_extended_datalog(atom))
        self.assertTrue(compile.is_extended_datalog(atom2))
        self.assertTrue(compile.is_extended_datalog(atom3))
        self.assertFalse(compile.is_extended_datalog(lit))
        self.assertTrue(compile.is_extended_datalog(regular_rule))
        self.assertTrue(compile.is_extended_datalog(regular_rule2))
        self.assertTrue(compile.is_extended_datalog(multi_rule))
        self.assertFalse(compile.is_extended_datalog(fake_rule))
        self.assertFalse(compile.is_extended_datalog(fake_rule2))
        self.assertFalse(compile.is_extended_datalog("a string"))

    def test_rule_validation(self):
        """Test that rules are properly validated."""
        # unsafe var in head
        rule = compile.parse1('p(x) :- q(y)')
        errs = compile.rule_errors(rule)
        self.assertEqual(len(errs), 1)

        # multiple unsafe vars in head
        rule = compile.parse1('p(x,y,z) :- q(w)')
        errs = compile.rule_errors(rule)
        self.assertEqual(len(set([str(x) for x in errs])), 3)

        # unsafe var in negtative literal:
        rule = compile.parse1('p(x) :- q(x), not r(y)')
        errs = compile.rule_errors(rule)
        self.assertEqual(len(set([str(x) for x in errs])), 1)

        # unsafe var in negative literal: ensure head doesn't make safe
        rule = compile.parse1('p(x) :- not q(x)')
        errs = compile.rule_errors(rule)
        self.assertEqual(len(set([str(x) for x in errs])), 1)

        # unsafe var in negative literal:
        #      ensure partial safety not total safety
        rule = compile.parse1('p(x) :- q(x), not r(x,y)')
        errs = compile.rule_errors(rule)
        self.assertEqual(len(set([str(x) for x in errs])), 1)

        # unsafe var in negative literal: ensure double negs doesn't make safe
        rule = compile.parse1('p(x) :- q(x), not r(x,y), not s(x, y)')
        errs = compile.rule_errors(rule)
        self.assertEqual(len(set([str(x) for x in errs])), 1)

        # multiple heads with modal
        rule = compile.parse1('execute[p(x)], r(x) :- q(x)')
        errs = compile.rule_errors(rule)
        self.assertEqual(len(set([str(x) for x in errs])), 1)

        # modal in body
        rule = compile.parse1('p(x) :- execute[q(x)]')
        errs = compile.rule_errors(rule)
        self.assertEqual(len(set([str(x) for x in errs])), 1)

        # keywords
        rule = compile.parse1('equal(x) :- q(x)')
        errs = compile.rule_errors(rule)
        self.assertEqual(len(set([str(x) for x in errs])), 1)

    def test_module_schemas(self):
        """Test that rules are properly checked against module schemas."""

        run = agnostic.Runtime()
        run.create_policy('mod1')
        run.create_policy('mod2')
        run.set_schema('mod1', compile.Schema({'p': (1, 2, 3), 'q': (1,)}),
                       complete=True)
        run.set_schema('mod2', compile.Schema({'p': (1,), 'q': (1, 2)}),
                       complete=True)

        def check_err(code_string, theory, emsg, msg, f=compile.rule_errors):
            rule = compile.parse1(code_string)
            errs = f(rule, run.theory, theory)
            self.assertTrue(any(emsg in str(err) for err in errs),
                            msg + ":: Failed to find error message '" + emsg +
                            "' in: " + ";".join(str(e) for e in errs))

        # no errors
        rule = compile.parse1('p(x) :- q(x), mod1:p(x, y, z), mod2:q(x, y), '
                              'mod1:q(t), mod2:p(t)')
        errs = compile.rule_errors(rule, run.theory)
        self.assertEqual(len(errs), 0, "Should not have found any errors")

        # unknown table within module
        check_err('p(x) :- q(x), mod1:r(x), r(x)',
                  'mod3',
                  'unknown table',
                  'Unknown table for rule')

        # wrong number of arguments
        check_err('p(x) :- q(x), mod1:p(x,y,z,w), r(x)',
                  'mod3',
                  'only 3 arguments are permitted',
                  'Wrong number of arguments for rule')

        # same tests for an atom

        # no errors
        atom = compile.parse1('p(1, 2, 2)')
        errs = compile.fact_errors(atom, run.theory, 'mod1')
        self.assertEqual(len(errs), 0, "Should not have found any errors")

        # unknown table within module
        check_err('r(1)',
                  'mod1',
                  'unknown table',
                  'Unknown table for atom',
                  f=compile.fact_errors)

        # wrong number of arguments
        check_err('p(1, 2, 3, 4)',
                  'mod1',
                  'only 3 arguments are permitted',
                  'Wrong number of arguments for atom',
                  f=compile.fact_errors)

        # schema update
        schema = compile.Schema()
        rule1 = compile.parse1('p(x) :- q(x, y)')
        change1 = schema.update(rule1.head, True)
        rule2 = compile.parse1('p(x) :- r(x, y)')
        change2 = schema.update(rule2.head, True)
        self.assertEqual(schema.count['p'], 2)
        schema.revert(change2)
        self.assertEqual(schema.count['p'], 1)
        schema.revert(change1)
        self.assertEqual('p' in schema.count, False)

        schema.update(rule1.head, True)
        schema.update(rule2.head, True)
        change1 = schema.update(rule1.head, False)
        change2 = schema.update(rule2.head, False)
        self.assertEqual('p' in schema.count, False)
        schema.revert(change2)
        self.assertEqual(schema.count['p'], 1)
        schema.revert(change1)
        self.assertEqual(schema.count['p'], 2)

    def test_rule_recursion(self):
        rules = compile.parse('p(x) :- q(x), r(x)  q(x) :- r(x) r(x) :- t(x)')
        self.assertFalse(compile.is_recursive(rules))

        rules = compile.parse('p(x) :- p(x)')
        self.assertTrue(compile.is_recursive(rules))

        rules = compile.parse('p(x) :- q(x)  q(x) :- r(x)  r(x) :- p(x)')
        self.assertTrue(compile.is_recursive(rules))

        rules = compile.parse('p(x) :- q(x)  q(x) :- not p(x)')
        self.assertTrue(compile.is_recursive(rules))

        rules = compile.parse('p(x) :- q(x), s(x)  q(x) :- t(x)  s(x) :- p(x)')
        self.assertTrue(compile.is_recursive(rules))

    def test_rule_stratification(self):
        rules = compile.parse('p(x) :- not q(x)')
        self.assertTrue(compile.is_stratified(rules))

        rules = compile.parse('p(x) :- p(x)')
        self.assertTrue(compile.is_stratified(rules))

        rules = compile.parse('p(x) :- q(x)  q(x) :- p(x)')
        self.assertTrue(compile.is_stratified(rules))

        rules = compile.parse('p(x) :- q(x)  q(x) :- not r(x)')
        self.assertTrue(compile.is_stratified(rules))

        rules = compile.parse('p(x) :- not q(x)  q(x) :- not r(x)')
        self.assertTrue(compile.is_stratified(rules))

        rules = compile.parse('p(x) :- not q(x)  '
                              'q(x) :- not r(x)  '
                              'r(x) :- not s(x)')
        self.assertTrue(compile.is_stratified(rules))

        rules = compile.parse('p(x) :- q(x), r(x) '
                              'q(x) :- not t(x) '
                              'r(x) :- not s(x)')
        self.assertTrue(compile.is_stratified(rules))

        rules = compile.parse('p(x) :- not p(x)')
        self.assertFalse(compile.is_stratified(rules))

        rules = compile.parse('p(x) :- q(x)  q(x) :- not p(x)')
        self.assertFalse(compile.is_stratified(rules))

        rules = compile.parse('p(x) :- q(x),r(x)  r(x) :- not p(x)')
        self.assertFalse(compile.is_stratified(rules))

        rules = compile.parse('p(x) :- q(x), r(x) '
                              'q(x) :- not t(x) '
                              'r(x) :- not s(x) '
                              't(x) :- p(x)')
        self.assertFalse(compile.is_stratified(rules))


class TestDependencyGraph(base.TestCase):

    def test_nodes_edges(self):
        g = compile.RuleDependencyGraph()

        # first insertion
        g.formula_insert(compile.parse1('p(x), q(x) :- r(x), s(x)'))
        self.assertTrue(g.node_in('p'))
        self.assertTrue(g.node_in('q'))
        self.assertTrue(g.node_in('r'))
        self.assertTrue(g.node_in('s'))
        self.assertTrue(g.edge_in('p', 'r', False))
        self.assertTrue(g.edge_in('p', 's', False))
        self.assertTrue(g.edge_in('q', 'r', False))
        self.assertTrue(g.edge_in('q', 's', False))
        self.assertFalse(g.has_cycle())

        # another insertion
        g.formula_insert(compile.parse1('r(x) :- t(x)'))
        self.assertTrue(g.node_in('p'))
        self.assertTrue(g.node_in('q'))
        self.assertTrue(g.node_in('r'))
        self.assertTrue(g.node_in('s'))
        self.assertTrue(g.edge_in('p', 'r', False))
        self.assertTrue(g.edge_in('p', 's', False))
        self.assertTrue(g.edge_in('q', 'r', False))
        self.assertTrue(g.edge_in('q', 's', False))
        self.assertTrue(g.node_in('t'))
        self.assertTrue(g.edge_in('r', 't', False))
        self.assertFalse(g.has_cycle())

        # 3rd insertion, creating a cycle
        g.formula_insert(compile.parse1('t(x) :- p(x)'))
        self.assertTrue(g.edge_in('t', 'p', False))
        self.assertTrue(g.has_cycle())

        # deletion
        g.formula_delete(compile.parse1('p(x), q(x) :- r(x), s(x)'))
        self.assertTrue(g.node_in('p'))
        self.assertTrue(g.node_in('r'))
        self.assertTrue(g.node_in('t'))
        self.assertTrue(g.edge_in('r', 't', False))
        self.assertTrue(g.edge_in('t', 'p', False))
        self.assertFalse(g.has_cycle())

        # double-insertion
        g.formula_insert(compile.parse1('p(x) :- q(x), r(x)'))
        g.formula_insert(compile.parse1('p(1) :- r(1)'))
        self.assertTrue(g.has_cycle())

        # deletion -- checking for bag semantics
        g.formula_delete(compile.parse1('p(1) :- r(1)'))
        self.assertTrue(g.has_cycle())
        g.formula_delete(compile.parse1('p(x) :- q(x), r(x)'))
        self.assertFalse(g.has_cycle())

        # update
        g.formula_update([
            compile.Event(compile.parse1('a(x) :- b(x)')),
            compile.Event(compile.parse1('b(x) :- c(x)')),
            compile.Event(compile.parse1('c(x) :- a(x)'))])
        self.assertTrue(g.has_cycle())
        g.formula_update([
            compile.Event(compile.parse1('c(x) :- a(x)'), insert=False)])
        self.assertFalse(g.has_cycle())

        # cycle enumeration
        g = compile.RuleDependencyGraph()
        g.formula_insert(compile.parse1('p(x) :- q(x), r(x)'))
        g.formula_insert(compile.parse1('q(x) :- t(x), not s(x)'))
        g.formula_insert(compile.parse1('t(x) :- t(x), p(x), q(x)'))
        self.assertTrue(g.has_cycle())
        self.assertEqual(len(g.cycles()), 3)
        expected_cycle_set = set([
            utility.Cycle(['p', 'q', 't', 'p']),
            utility.Cycle(['q', 't', 'q']),
            utility.Cycle(['t', 't'])
        ])
        actual_cycle_set = set([
            utility.Cycle(g.cycles()[0]),
            utility.Cycle(g.cycles()[1]),
            utility.Cycle(g.cycles()[2])
        ])
        self.assertEqual(expected_cycle_set, actual_cycle_set)

    def test_dependencies(self):
        g = compile.RuleDependencyGraph()
        g.formula_insert(compile.parse1('p(x) :- q(x), r(x)'))
        g.formula_insert(compile.parse1('q(x) :- t(x), not s(x)'))
        self.assertEqual(g.dependencies('p'), set(['p', 'q', 'r', 't', 's']))
        self.assertEqual(g.dependencies('q'), set(['q', 't', 's']))
        self.assertEqual(g.dependencies('r'), set(['r']))
        self.assertEqual(g.dependencies('t'), set(['t']))
        self.assertEqual(g.dependencies('s'), set(['s']))

        # cyclic case
        g = compile.RuleDependencyGraph()
        g.formula_insert(compile.parse1('p(x) :- q(x), r(x)'))
        g.formula_insert(compile.parse1('q(x) :- t(x), not s(x)'))
        g.formula_insert(compile.parse1('t(x) :- t(x), p(x), q(x)'))
        self.assertEqual(g.dependencies('p'), set(['p', 'q', 'r', 't', 's']))
        self.assertEqual(g.dependencies('q'), set(['p', 'q', 'r', 't', 's']))
        self.assertEqual(g.dependencies('r'), set(['r']))
        self.assertEqual(g.dependencies('t'), set(['p', 'q', 'r', 't', 's']))
        self.assertEqual(g.dependencies('s'), set(['s']))

        g = compile.RuleDependencyGraph(head_to_body=False)
        g.formula_insert(compile.parse1('p(x) :- q(x), r(x)'))
        g.formula_insert(compile.parse1('q(x) :- t(x), not s(x)'))
        self.assertEqual(g.dependencies('p'), set(['p']))
        self.assertEqual(g.dependencies('q'), set(['q', 'p']))
        self.assertEqual(g.dependencies('r'), set(['r', 'p']))
        self.assertEqual(g.dependencies('t'), set(['t', 'q', 'p']))
        self.assertEqual(g.dependencies('s'), set(['s', 'q', 'p']))

    def test_modal_index(self):
        m = analysis.ModalIndex()
        m.add('execute', 'p')
        self.assertEqual(set(m.tables('execute')), set(['p']))
        m.add('execute', 'q')
        self.assertEqual(set(m.tables('execute')), set(['p', 'q']))
        m.remove('execute', 'q')
        self.assertEqual(set(m.tables('execute')), set(['p']))
        m.add('execute', 'q')
        m.add('execute', 'q')
        m.remove('execute', 'q')
        self.assertEqual(set(m.tables('execute')), set(['p', 'q']))
        m.remove('execute', 'q')
        self.assertEqual(set(m.tables('execute')), set(['p']))
        m.add('foo', 'p')
        self.assertEqual(set(m.tables('foo')), set(['p']))
        self.assertEqual(set(m.tables('bar')), set())
        self.assertEqual(set(m.tables('execute')), set(['p']))

    def test_modal_index_composition(self):
        m = analysis.ModalIndex()
        m.add('execute', 'p')
        m.add('execute', 'q')
        m.add('execute', 'r')
        m.add('foo', 'r')
        m.add('foo', 's')

        n = analysis.ModalIndex()
        n.add('execute', 'p')
        n.add('execute', 'alpha')
        n.add('foo', 'r')
        n.add('bar', 'beta')

        n_plus_m = analysis.ModalIndex()
        n_plus_m.add('execute', 'p')
        n_plus_m.add('execute', 'p')
        n_plus_m.add('execute', 'q')
        n_plus_m.add('execute', 'r')
        n_plus_m.add('execute', 'alpha')
        n_plus_m.add('foo', 'r')
        n_plus_m.add('foo', 's')
        n_plus_m.add('foo', 'r')
        n_plus_m.add('bar', 'beta')

        m_copy = copy.copy(m)
        m_copy += n
        self.assertEqual(m_copy, n_plus_m)

        m_minus_n = analysis.ModalIndex()
        m_minus_n.add('execute', 'q')
        m_minus_n.add('execute', 'r')
        m_minus_n.add('foo', 's')

        m_copy = copy.copy(m)
        m_copy -= n
        self.assertEqual(m_copy, m_minus_n)

    def test_modals(self):
        g = compile.RuleDependencyGraph()
        g.formula_insert(compile.parse1('p(x) :- q(x)'))
        g.formula_insert(compile.parse1('q(x) :- r(x)'))
        g.formula_insert(compile.parse1('execute[p(x)] :- q(x)'))
        chgs = g.formula_insert(compile.parse1('execute[r(x)] :- q(x)'))
        g.formula_insert(compile.parse1('insert[s(x)] :- q(x)'))
        self.assertEqual(set(g.tables_with_modal('execute')), set(['p', 'r']))
        g.undo_changes(chgs)
        self.assertEqual(set(g.tables_with_modal('execute')), set(['p']))
        chgs = g.formula_delete(compile.parse1('execute[p(x)] :- q(x)'))
        self.assertEqual(set(g.tables_with_modal('execute')), set())
        g.undo_changes(chgs)
        self.assertEqual(set(g.tables_with_modal('execute')), set(['p']))



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from oslo_log import log as logging

from congress.api import api_utils
from congress.api import webservice
from congress.dse import deepsix
from congress import exception


LOG = logging.getLogger(__name__)


def d6service(name, keys, inbox, datapath, args):
    return DatasourceDriverModel(name, keys, inbox=inbox,
                                 dataPath=datapath, **args)


class DatasourceDriverModel(deepsix.deepSix):
    """Model for handling API requests about DatasourceDriver."""
    def __init__(self, name, keys, inbox=None, dataPath=None,
                 datasource_mgr=None):
        super(DatasourceDriverModel, self).__init__(name, keys, inbox=inbox,
                                                    dataPath=dataPath)
        self.datasource_mgr = datasource_mgr

    def rpc(self, caller, name, *args, **kwargs):
        f = getattr(caller, name)
        return f(*args, **kwargs)

    def get_items(self, params, context=None):
        """Get items in model.

        Args:
            params: A dict-like object containing parameters
                    from the request query string and body.
            context: Key-values providing frame of reference of request

        Returns: A dict containing at least a 'results' key whose value is
                 a list of items in the model.  Additional keys set in the
                 dict will also be rendered for the user.
        """
        drivers = self.rpc(self.datasource_mgr, 'get_drivers_info')
        fields = ['id', 'description']
        results = [self.datasource_mgr.make_datasource_dict(
                   driver, fields=fields)
                   for driver in drivers]
        return {"results": results}

    def get_item(self, id_, params, context=None):
        """Retrieve item with id id_ from model.

        Args:
            id_: The ID of the item to retrieve
            params: A dict-like object containing parameters
                    from the request query string and body.
            context: Key-values providing frame of reference of request

        Returns:
             The matching item or None if item with id_ does not exist.
        """
        datasource = context.get('driver_id')
        try:
            schema = self.rpc(self.datasource_mgr, 'get_driver_schema',
                              datasource)
            driver = self.rpc(self.datasource_mgr, 'get_driver_info',
                              datasource)
        except exception.DriverNotFound as e:
            raise webservice.DataModelException(e.code, str(e),
                                                http_status_code=e.code)

        tables = [api_utils.create_table_dict(table_, schema)
                  for table_ in schema]
        driver['tables'] = tables
        return driver
